// minimal browser mocks needed for some nodejs modules
      if (window) {
        window.global = window
        window.process = {
          nextTick: function(cb) {
            new Promise(() => cb && cb())
          }
        }
      }
    

import randomBytes from 'randombytes';
import EventSource from 'eventsource';
import require$$0$1 from 'events';
import fetch from 'isomorphic-fetch';
import require$$0 from 'debug';
import isBuffer from 'is-buffer';
import base32Encode from 'base32-encode';
import prettyBytes from 'pretty-bytes';
import require$$0$2 from 'url';

const debug$2 = require$$0('sonar:fetch');


/**
  * Fetch a resource.
  *
  * This is a wrapper around the fetch web API. It should be API compatible to fetch,
  * with the following changes:
  *
  * @async
  * @param {string} [opts.endpoint=''] Endpoint URL (will be prefixed to URL)
  * @param {string} [opts.requestType='json'] Request encoding and content type.
  *   Supported values are 'json' and 'binary'
  * @param {string} [opts.responseType='text'] Response encoding. If the response
  *    has a JSON content type, will always be set to 'json'.
  *    Supported values are 'text', 'binary' and 'stream'.
  * @param {object} [opts.params] Query string parameters (will be encoded correctly).
  *
  * @return {Promise<object>} If the response has a JSON content type header, the
  *    decoded JSON will be returned. if opts.responseType is 'binary' or 'text',
  *    the response will be returned as a buffer or text.
  *
  * TODO: Rethink the default responseType cascade.
  */
var fetch_1 = async function makeFetch (url, opts) {
  if (!url.match(/^https?:\/\//)) {
    if (url.indexOf('://') !== -1) throw new Error('Only http: and https: protocols are supported.')
    if (!url.startsWith('/')) url = '/' + url;
    if (opts.endpoint) url = opts.endpoint + url;
  }

  if (!opts.headers) opts.headers = {};
  if (!opts.requestType) {
    if (isBuffer(opts.body)) opts.requestType = 'buffer';
    else opts.requestType = 'json';
  }

  if (opts.params) {
    const searchParams = new URLSearchParams();
    for (const [key, value] of Object.entries(opts.params)) {
      searchParams.append(key, value);
    }
    url += '?' + searchParams.toString();
  }

  if (opts.requestType === 'json') {
    opts.body = JSON.stringify(opts.body);
    opts.headers['content-type'] = 'application/json';
  }
  if (opts.requestType === 'buffer') {
    opts.headers['content-type'] = 'application/octet-stream';
  }

  const log = opts.log === false ? () => {} : (opts.log || debug$2);

  try {
    debug$2('fetch', url, opts);
    const res = await fetch(url, opts);
    if (!res.ok) {
      let message;
      if (isJsonResponse(res)) {
        message = (await res.json()).error;
      } else {
        message = await res.text();
      }
      log(`error ${res.status} ${url}`);
      throw new Error('Remote error (code ' + res.status + '): ' + message)
    }

    log(`ok ${res.status} fetch ${url}`);

    if (opts.responseType === 'stream') {
      return res.body
    }
    if (opts.responseType === 'buffer') {
      // nodejs only: res.buffer() returns a Buffer instance.
      if (res.buffer) return await res.buffer()
      // browser: Fetch API res.arrayBuffer returns ArrayBuffer.
      else return await res.arrayBuffer()
    }

    if (isJsonResponse(res)) {
      return await res.json()
    }

    return await res.text()
  } catch (err) {
    // TODO: If error fails for insufficient authorization, try creating
    // a new token if accessCode is set
    debug$2('fetch error', err);
    log(`error: fetch ${url} ${err.message}`);
    throw err
  }
};

function isJsonResponse (res) {
  const header = res.headers.get('content-type');
  if (!header) return false
  return header.indexOf('application/json') !== -1
}

var queueMicrotask_1 = typeof queueMicrotask === 'function' ? queueMicrotask : (fn) => Promise.resolve().then(fn);

var fixedSize = class FixedFIFO {
  constructor (hwm) {
    if (!(hwm > 0) || ((hwm - 1) & hwm) !== 0) throw new Error('Max size for a FixedFIFO should be a power of two')
    this.buffer = new Array(hwm);
    this.mask = hwm - 1;
    this.top = 0;
    this.btm = 0;
    this.next = null;
  }

  push (data) {
    if (this.buffer[this.top] !== undefined) return false
    this.buffer[this.top] = data;
    this.top = (this.top + 1) & this.mask;
    return true
  }

  shift () {
    const last = this.buffer[this.btm];
    if (last === undefined) return undefined
    this.buffer[this.btm] = undefined;
    this.btm = (this.btm + 1) & this.mask;
    return last
  }

  isEmpty () {
    return this.buffer[this.btm] === undefined
  }
};

var fastFifo = class FastFIFO {
  constructor (hwm) {
    this.hwm = hwm || 16;
    this.head = new fixedSize(this.hwm);
    this.tail = this.head;
  }

  push (val) {
    if (!this.head.push(val)) {
      const prev = this.head;
      this.head = prev.next = new fixedSize(2 * this.head.buffer.length);
      this.head.push(val);
    }
  }

  shift () {
    const val = this.tail.shift();
    if (val === undefined && this.tail.next) {
      const next = this.tail.next;
      this.tail.next = null;
      this.tail = next;
      return this.tail.shift()
    }
    return val
  }

  isEmpty () {
    return this.head.isEmpty()
  }
};

const { EventEmitter: EventEmitter$2 } = require$$0$1;
const STREAM_DESTROYED = new Error('Stream was destroyed');
const PREMATURE_CLOSE = new Error('Premature close');




/* eslint-disable no-multi-spaces */

const MAX = ((1 << 25) - 1);

// Shared state
const OPENING     = 0b001;
const DESTROYING  = 0b010;
const DESTROYED   = 0b100;

const NOT_OPENING = MAX ^ OPENING;

// Read state
const READ_ACTIVE           = 0b0000000000001 << 3;
const READ_PRIMARY          = 0b0000000000010 << 3;
const READ_SYNC             = 0b0000000000100 << 3;
const READ_QUEUED           = 0b0000000001000 << 3;
const READ_RESUMED          = 0b0000000010000 << 3;
const READ_PIPE_DRAINED     = 0b0000000100000 << 3;
const READ_ENDING           = 0b0000001000000 << 3;
const READ_EMIT_DATA        = 0b0000010000000 << 3;
const READ_EMIT_READABLE    = 0b0000100000000 << 3;
const READ_EMITTED_READABLE = 0b0001000000000 << 3;
const READ_DONE             = 0b0010000000000 << 3;
const READ_NEXT_TICK        = 0b0100000000001 << 3; // also active
const READ_NEEDS_PUSH       = 0b1000000000000 << 3;

const READ_NOT_ACTIVE             = MAX ^ READ_ACTIVE;
const READ_NON_PRIMARY            = MAX ^ READ_PRIMARY;
const READ_NON_PRIMARY_AND_PUSHED = MAX ^ (READ_PRIMARY | READ_NEEDS_PUSH);
const READ_NOT_SYNC               = MAX ^ READ_SYNC;
const READ_PUSHED                 = MAX ^ READ_NEEDS_PUSH;
const READ_PAUSED                 = MAX ^ READ_RESUMED;
const READ_NOT_QUEUED             = MAX ^ (READ_QUEUED | READ_EMITTED_READABLE);
const READ_NOT_ENDING             = MAX ^ READ_ENDING;
const READ_PIPE_NOT_DRAINED       = MAX ^ (READ_RESUMED | READ_PIPE_DRAINED);
const READ_NOT_NEXT_TICK          = MAX ^ READ_NEXT_TICK;

// Write state
const WRITE_ACTIVE     = 0b000000001 << 16;
const WRITE_PRIMARY    = 0b000000010 << 16;
const WRITE_SYNC       = 0b000000100 << 16;
const WRITE_QUEUED     = 0b000001000 << 16;
const WRITE_UNDRAINED  = 0b000010000 << 16;
const WRITE_DONE       = 0b000100000 << 16;
const WRITE_EMIT_DRAIN = 0b001000000 << 16;
const WRITE_NEXT_TICK  = 0b010000001 << 16; // also active
const WRITE_FINISHING  = 0b100000000 << 16;

const WRITE_NOT_ACTIVE    = MAX ^ WRITE_ACTIVE;
const WRITE_NOT_SYNC      = MAX ^ WRITE_SYNC;
const WRITE_NON_PRIMARY   = MAX ^ WRITE_PRIMARY;
const WRITE_NOT_FINISHING = MAX ^ WRITE_FINISHING;
const WRITE_DRAINED       = MAX ^ WRITE_UNDRAINED;
const WRITE_NOT_QUEUED    = MAX ^ WRITE_QUEUED;
const WRITE_NOT_NEXT_TICK = MAX ^ WRITE_NEXT_TICK;

// Combined shared state
const ACTIVE = READ_ACTIVE | WRITE_ACTIVE;
const NOT_ACTIVE = MAX ^ ACTIVE;
const DONE = READ_DONE | WRITE_DONE;
const DESTROY_STATUS = DESTROYING | DESTROYED;
const OPEN_STATUS = DESTROY_STATUS | OPENING;
const AUTO_DESTROY = DESTROY_STATUS | DONE;
const NON_PRIMARY = WRITE_NON_PRIMARY & READ_NON_PRIMARY;
const TICKING = (WRITE_NEXT_TICK | READ_NEXT_TICK) & NOT_ACTIVE;
const ACTIVE_OR_TICKING = ACTIVE | TICKING;
const IS_OPENING = OPEN_STATUS | TICKING;

// Combined read state
const READ_PRIMARY_STATUS = OPEN_STATUS | READ_ENDING | READ_DONE;
const READ_STATUS = OPEN_STATUS | READ_DONE | READ_QUEUED;
const READ_FLOWING = READ_RESUMED | READ_PIPE_DRAINED;
const READ_ACTIVE_AND_SYNC = READ_ACTIVE | READ_SYNC;
const READ_ACTIVE_AND_SYNC_AND_NEEDS_PUSH = READ_ACTIVE | READ_SYNC | READ_NEEDS_PUSH;
const READ_PRIMARY_AND_ACTIVE = READ_PRIMARY | READ_ACTIVE;
const READ_ENDING_STATUS = OPEN_STATUS | READ_ENDING | READ_QUEUED;
const READ_EMIT_READABLE_AND_QUEUED = READ_EMIT_READABLE | READ_QUEUED;
const READ_READABLE_STATUS = OPEN_STATUS | READ_EMIT_READABLE | READ_QUEUED | READ_EMITTED_READABLE;
const SHOULD_NOT_READ = OPEN_STATUS | READ_ACTIVE | READ_ENDING | READ_DONE | READ_NEEDS_PUSH;
const READ_BACKPRESSURE_STATUS = DESTROY_STATUS | READ_ENDING | READ_DONE;

// Combined write state
const WRITE_PRIMARY_STATUS = OPEN_STATUS | WRITE_FINISHING | WRITE_DONE;
const WRITE_QUEUED_AND_UNDRAINED = WRITE_QUEUED | WRITE_UNDRAINED;
const WRITE_QUEUED_AND_ACTIVE = WRITE_QUEUED | WRITE_ACTIVE;
const WRITE_DRAIN_STATUS = WRITE_QUEUED | WRITE_UNDRAINED | OPEN_STATUS | WRITE_ACTIVE;
const WRITE_STATUS = OPEN_STATUS | WRITE_ACTIVE | WRITE_QUEUED;
const WRITE_PRIMARY_AND_ACTIVE = WRITE_PRIMARY | WRITE_ACTIVE;
const WRITE_ACTIVE_AND_SYNC = WRITE_ACTIVE | WRITE_SYNC;
const WRITE_FINISHING_STATUS = OPEN_STATUS | WRITE_FINISHING | WRITE_QUEUED;
const WRITE_BACKPRESSURE_STATUS = WRITE_UNDRAINED | DESTROY_STATUS | WRITE_FINISHING | WRITE_DONE;

const asyncIterator = Symbol.asyncIterator || Symbol('asyncIterator');

class WritableState {
  constructor (stream, { highWaterMark = 16384, map = null, mapWritable, byteLength, byteLengthWritable } = {}) {
    this.stream = stream;
    this.queue = new fastFifo();
    this.highWaterMark = highWaterMark;
    this.buffered = 0;
    this.error = null;
    this.pipeline = null;
    this.byteLength = byteLengthWritable || byteLength || defaultByteLength;
    this.map = mapWritable || map;
    this.afterWrite = afterWrite.bind(this);
    this.afterUpdateNextTick = updateWriteNT.bind(this);
  }

  get ended () {
    return (this.stream._duplexState & WRITE_DONE) !== 0
  }

  push (data) {
    if (this.map !== null) data = this.map(data);

    this.buffered += this.byteLength(data);
    this.queue.push(data);

    if (this.buffered < this.highWaterMark) {
      this.stream._duplexState |= WRITE_QUEUED;
      return true
    }

    this.stream._duplexState |= WRITE_QUEUED_AND_UNDRAINED;
    return false
  }

  shift () {
    const data = this.queue.shift();
    const stream = this.stream;

    this.buffered -= this.byteLength(data);
    if (this.buffered === 0) stream._duplexState &= WRITE_NOT_QUEUED;

    return data
  }

  end (data) {
    if (typeof data === 'function') this.stream.once('finish', data);
    else if (data !== undefined && data !== null) this.push(data);
    this.stream._duplexState = (this.stream._duplexState | WRITE_FINISHING) & WRITE_NON_PRIMARY;
  }

  autoBatch (data, cb) {
    const buffer = [];
    const stream = this.stream;

    buffer.push(data);
    while ((stream._duplexState & WRITE_STATUS) === WRITE_QUEUED_AND_ACTIVE) {
      buffer.push(stream._writableState.shift());
    }

    if ((stream._duplexState & OPEN_STATUS) !== 0) return cb(null)
    stream._writev(buffer, cb);
  }

  update () {
    const stream = this.stream;

    while ((stream._duplexState & WRITE_STATUS) === WRITE_QUEUED) {
      const data = this.shift();
      stream._duplexState |= WRITE_ACTIVE_AND_SYNC;
      stream._write(data, this.afterWrite);
      stream._duplexState &= WRITE_NOT_SYNC;
    }

    if ((stream._duplexState & WRITE_PRIMARY_AND_ACTIVE) === 0) this.updateNonPrimary();
  }

  updateNonPrimary () {
    const stream = this.stream;

    if ((stream._duplexState & WRITE_FINISHING_STATUS) === WRITE_FINISHING) {
      stream._duplexState = (stream._duplexState | WRITE_ACTIVE) & WRITE_NOT_FINISHING;
      stream._final(afterFinal.bind(this));
      return
    }

    if ((stream._duplexState & DESTROY_STATUS) === DESTROYING) {
      if ((stream._duplexState & ACTIVE_OR_TICKING) === 0) {
        stream._duplexState |= ACTIVE;
        stream._destroy(afterDestroy.bind(this));
      }
      return
    }

    if ((stream._duplexState & IS_OPENING) === OPENING) {
      stream._duplexState = (stream._duplexState | ACTIVE) & NOT_OPENING;
      stream._open(afterOpen.bind(this));
    }
  }

  updateNextTick () {
    if ((this.stream._duplexState & WRITE_NEXT_TICK) !== 0) return
    this.stream._duplexState |= WRITE_NEXT_TICK;
    queueMicrotask_1(this.afterUpdateNextTick);
  }
}

class ReadableState {
  constructor (stream, { highWaterMark = 16384, map = null, mapReadable, byteLength, byteLengthReadable } = {}) {
    this.stream = stream;
    this.queue = new fastFifo();
    this.highWaterMark = highWaterMark;
    this.buffered = 0;
    this.error = null;
    this.pipeline = null;
    this.byteLength = byteLengthReadable || byteLength || defaultByteLength;
    this.map = mapReadable || map;
    this.pipeTo = null;
    this.afterRead = afterRead.bind(this);
    this.afterUpdateNextTick = updateReadNT.bind(this);
  }

  get ended () {
    return (this.stream._duplexState & READ_DONE) !== 0
  }

  pipe (pipeTo, cb) {
    if (this.pipeTo !== null) throw new Error('Can only pipe to one destination')

    this.stream._duplexState |= READ_PIPE_DRAINED;
    this.pipeTo = pipeTo;
    this.pipeline = new Pipeline(this.stream, pipeTo, cb || null);

    if (cb) this.stream.on('error', noop$2); // We already error handle this so supress crashes

    if (isStreamx(pipeTo)) {
      pipeTo._writableState.pipeline = this.pipeline;
      if (cb) pipeTo.on('error', noop$2); // We already error handle this so supress crashes
      pipeTo.on('finish', this.pipeline.finished.bind(this.pipeline)); // TODO: just call finished from pipeTo itself
    } else {
      const onerror = this.pipeline.done.bind(this.pipeline, pipeTo);
      const onclose = this.pipeline.done.bind(this.pipeline, pipeTo, null); // onclose has a weird bool arg
      pipeTo.on('error', onerror);
      pipeTo.on('close', onclose);
      pipeTo.on('finish', this.pipeline.finished.bind(this.pipeline));
    }

    pipeTo.on('drain', afterDrain.bind(this));
    this.stream.emit('piping', pipeTo);
    pipeTo.emit('pipe', this.stream);
  }

  push (data) {
    const stream = this.stream;

    if (data === null) {
      this.highWaterMark = 0;
      stream._duplexState = (stream._duplexState | READ_ENDING) & READ_NON_PRIMARY_AND_PUSHED;
      return false
    }

    if (this.map !== null) data = this.map(data);
    this.buffered += this.byteLength(data);
    this.queue.push(data);

    stream._duplexState = (stream._duplexState | READ_QUEUED) & READ_PUSHED;

    return this.buffered < this.highWaterMark
  }

  shift () {
    const data = this.queue.shift();

    this.buffered -= this.byteLength(data);
    if (this.buffered === 0) this.stream._duplexState &= READ_NOT_QUEUED;
    return data
  }

  unshift (data) {
    let tail;
    const pending = [];

    while ((tail = this.queue.shift()) !== undefined) {
      pending.push(tail);
    }

    this.push(data);

    for (let i = 0; i < pending.length; i++) {
      this.queue.push(pending[i]);
    }
  }

  read () {
    const stream = this.stream;

    if ((stream._duplexState & READ_STATUS) === READ_QUEUED) {
      const data = this.shift();
      if (this.pipeTo !== null && this.pipeTo.write(data) === false) stream._duplexState &= READ_PIPE_NOT_DRAINED;
      if ((stream._duplexState & READ_EMIT_DATA) !== 0) stream.emit('data', data);
      return data
    }

    return null
  }

  drain () {
    const stream = this.stream;

    while ((stream._duplexState & READ_STATUS) === READ_QUEUED && (stream._duplexState & READ_FLOWING) !== 0) {
      const data = this.shift();
      if (this.pipeTo !== null && this.pipeTo.write(data) === false) stream._duplexState &= READ_PIPE_NOT_DRAINED;
      if ((stream._duplexState & READ_EMIT_DATA) !== 0) stream.emit('data', data);
    }
  }

  update () {
    const stream = this.stream;

    this.drain();

    while (this.buffered < this.highWaterMark && (stream._duplexState & SHOULD_NOT_READ) === 0) {
      stream._duplexState |= READ_ACTIVE_AND_SYNC_AND_NEEDS_PUSH;
      stream._read(this.afterRead);
      stream._duplexState &= READ_NOT_SYNC;
      if ((stream._duplexState & READ_ACTIVE) === 0) this.drain();
    }

    if ((stream._duplexState & READ_READABLE_STATUS) === READ_EMIT_READABLE_AND_QUEUED) {
      stream._duplexState |= READ_EMITTED_READABLE;
      stream.emit('readable');
    }

    if ((stream._duplexState & READ_PRIMARY_AND_ACTIVE) === 0) this.updateNonPrimary();
  }

  updateNonPrimary () {
    const stream = this.stream;

    if ((stream._duplexState & READ_ENDING_STATUS) === READ_ENDING) {
      stream._duplexState = (stream._duplexState | READ_DONE) & READ_NOT_ENDING;
      stream.emit('end');
      if ((stream._duplexState & AUTO_DESTROY) === DONE) stream._duplexState |= DESTROYING;
      if (this.pipeTo !== null) this.pipeTo.end();
    }

    if ((stream._duplexState & DESTROY_STATUS) === DESTROYING) {
      if ((stream._duplexState & ACTIVE_OR_TICKING) === 0) {
        stream._duplexState |= ACTIVE;
        stream._destroy(afterDestroy.bind(this));
      }
      return
    }

    if ((stream._duplexState & IS_OPENING) === OPENING) {
      stream._duplexState = (stream._duplexState | ACTIVE) & NOT_OPENING;
      stream._open(afterOpen.bind(this));
    }
  }

  updateNextTick () {
    if ((this.stream._duplexState & READ_NEXT_TICK) !== 0) return
    this.stream._duplexState |= READ_NEXT_TICK;
    queueMicrotask_1(this.afterUpdateNextTick);
  }
}

class TransformState {
  constructor (stream) {
    this.data = null;
    this.afterTransform = afterTransform.bind(stream);
    this.afterFinal = null;
  }
}

class Pipeline {
  constructor (src, dst, cb) {
    this.from = src;
    this.to = dst;
    this.afterPipe = cb;
    this.error = null;
    this.pipeToFinished = false;
  }

  finished () {
    this.pipeToFinished = true;
  }

  done (stream, err) {
    if (err) this.error = err;

    if (stream === this.to) {
      this.to = null;

      if (this.from !== null) {
        if ((this.from._duplexState & READ_DONE) === 0 || !this.pipeToFinished) {
          this.from.destroy(this.error || new Error('Writable stream closed prematurely'));
        }
        return
      }
    }

    if (stream === this.from) {
      this.from = null;

      if (this.to !== null) {
        if ((stream._duplexState & READ_DONE) === 0) {
          this.to.destroy(this.error || new Error('Readable stream closed before ending'));
        }
        return
      }
    }

    if (this.afterPipe !== null) this.afterPipe(this.error);
    this.to = this.from = this.afterPipe = null;
  }
}

function afterDrain () {
  this.stream._duplexState |= READ_PIPE_DRAINED;
  if ((this.stream._duplexState & READ_ACTIVE_AND_SYNC) === 0) this.updateNextTick();
}

function afterFinal (err) {
  const stream = this.stream;
  if (err) stream.destroy(err);
  if ((stream._duplexState & DESTROY_STATUS) === 0) {
    stream._duplexState |= WRITE_DONE;
    stream.emit('finish');
  }
  if ((stream._duplexState & AUTO_DESTROY) === DONE) {
    stream._duplexState |= DESTROYING;
  }

  stream._duplexState &= WRITE_NOT_ACTIVE;
  this.update();
}

function afterDestroy (err) {
  const stream = this.stream;

  if (!err && this.error !== STREAM_DESTROYED) err = this.error;
  if (err) stream.emit('error', err);
  stream._duplexState |= DESTROYED;
  stream.emit('close');

  const rs = stream._readableState;
  const ws = stream._writableState;

  if (rs !== null && rs.pipeline !== null) rs.pipeline.done(stream, err);
  if (ws !== null && ws.pipeline !== null) ws.pipeline.done(stream, err);
}

function afterWrite (err) {
  const stream = this.stream;

  if (err) stream.destroy(err);
  stream._duplexState &= WRITE_NOT_ACTIVE;

  if ((stream._duplexState & WRITE_DRAIN_STATUS) === WRITE_UNDRAINED) {
    stream._duplexState &= WRITE_DRAINED;
    if ((stream._duplexState & WRITE_EMIT_DRAIN) === WRITE_EMIT_DRAIN) {
      stream.emit('drain');
    }
  }

  if ((stream._duplexState & WRITE_SYNC) === 0) this.update();
}

function afterRead (err) {
  if (err) this.stream.destroy(err);
  this.stream._duplexState &= READ_NOT_ACTIVE;
  if ((this.stream._duplexState & READ_SYNC) === 0) this.update();
}

function updateReadNT () {
  this.stream._duplexState &= READ_NOT_NEXT_TICK;
  this.update();
}

function updateWriteNT () {
  this.stream._duplexState &= WRITE_NOT_NEXT_TICK;
  this.update();
}

function afterOpen (err) {
  const stream = this.stream;

  if (err) stream.destroy(err);

  if ((stream._duplexState & DESTROYING) === 0) {
    if ((stream._duplexState & READ_PRIMARY_STATUS) === 0) stream._duplexState |= READ_PRIMARY;
    if ((stream._duplexState & WRITE_PRIMARY_STATUS) === 0) stream._duplexState |= WRITE_PRIMARY;
    stream.emit('open');
  }

  stream._duplexState &= NOT_ACTIVE;

  if (stream._writableState !== null) {
    stream._writableState.update();
  }

  if (stream._readableState !== null) {
    stream._readableState.update();
  }
}

function afterTransform (err, data) {
  if (data !== undefined && data !== null) this.push(data);
  this._writableState.afterWrite(err);
}

class Stream extends EventEmitter$2 {
  constructor (opts) {
    super();

    this._duplexState = 0;
    this._readableState = null;
    this._writableState = null;

    if (opts) {
      if (opts.open) this._open = opts.open;
      if (opts.destroy) this._destroy = opts.destroy;
      if (opts.predestroy) this._predestroy = opts.predestroy;
      if (opts.signal) {
        opts.signal.addEventListener('abort', abort.bind(this));
      }
    }
  }

  _open (cb) {
    cb(null);
  }

  _destroy (cb) {
    cb(null);
  }

  _predestroy () {
    // does nothing
  }

  get readable () {
    return this._readableState !== null ? true : undefined
  }

  get writable () {
    return this._writableState !== null ? true : undefined
  }

  get destroyed () {
    return (this._duplexState & DESTROYED) !== 0
  }

  get destroying () {
    return (this._duplexState & DESTROY_STATUS) !== 0
  }

  destroy (err) {
    if ((this._duplexState & DESTROY_STATUS) === 0) {
      if (!err) err = STREAM_DESTROYED;
      this._duplexState = (this._duplexState | DESTROYING) & NON_PRIMARY;
      if (this._readableState !== null) {
        this._readableState.error = err;
        this._readableState.updateNextTick();
      }
      if (this._writableState !== null) {
        this._writableState.error = err;
        this._writableState.updateNextTick();
      }
      this._predestroy();
    }
  }

  on (name, fn) {
    if (this._readableState !== null) {
      if (name === 'data') {
        this._duplexState |= (READ_EMIT_DATA | READ_RESUMED);
        this._readableState.updateNextTick();
      }
      if (name === 'readable') {
        this._duplexState |= READ_EMIT_READABLE;
        this._readableState.updateNextTick();
      }
    }

    if (this._writableState !== null) {
      if (name === 'drain') {
        this._duplexState |= WRITE_EMIT_DRAIN;
        this._writableState.updateNextTick();
      }
    }

    return super.on(name, fn)
  }
}

class Readable$1 extends Stream {
  constructor (opts) {
    super(opts);

    this._duplexState |= OPENING | WRITE_DONE;
    this._readableState = new ReadableState(this, opts);

    if (opts) {
      if (opts.read) this._read = opts.read;
    }
  }

  _read (cb) {
    cb(null);
  }

  pipe (dest, cb) {
    this._readableState.pipe(dest, cb);
    this._readableState.updateNextTick();
    return dest
  }

  read () {
    this._readableState.updateNextTick();
    return this._readableState.read()
  }

  push (data) {
    this._readableState.updateNextTick();
    return this._readableState.push(data)
  }

  unshift (data) {
    this._readableState.updateNextTick();
    return this._readableState.unshift(data)
  }

  resume () {
    this._duplexState |= READ_RESUMED;
    this._readableState.updateNextTick();
    return this
  }

  pause () {
    this._duplexState &= READ_PAUSED;
    return this
  }

  static _fromAsyncIterator (ite, opts) {
    let destroy;

    const rs = new Readable$1({
      ...opts,
      read (cb) {
        ite.next().then(push).then(cb.bind(null, null)).catch(cb);
      },
      predestroy () {
        destroy = ite.return();
      },
      destroy (cb) {
        destroy.then(cb.bind(null, null)).catch(cb);
      }
    });

    return rs

    function push (data) {
      if (data.done) rs.push(null);
      else rs.push(data.value);
    }
  }

  static from (data, opts) {
    if (isReadStreamx(data)) return data
    if (data[asyncIterator]) return this._fromAsyncIterator(data[asyncIterator](), opts)
    if (!Array.isArray(data)) data = data === undefined ? [] : [data];

    let i = 0;
    return new Readable$1({
      ...opts,
      read (cb) {
        this.push(i === data.length ? null : data[i++]);
        cb(null);
      }
    })
  }

  static isBackpressured (rs) {
    return (rs._duplexState & READ_BACKPRESSURE_STATUS) !== 0 || rs._readableState.buffered >= rs._readableState.highWaterMark
  }

  static isPaused (rs) {
    return (rs._duplexState & READ_RESUMED) === 0
  }

  [asyncIterator] () {
    const stream = this;

    let error = null;
    let promiseResolve = null;
    let promiseReject = null;

    this.on('error', (err) => { error = err; });
    this.on('readable', onreadable);
    this.on('close', onclose);

    return {
      [asyncIterator] () {
        return this
      },
      next () {
        return new Promise(function (resolve, reject) {
          promiseResolve = resolve;
          promiseReject = reject;
          const data = stream.read();
          if (data !== null) ondata(data);
          else if ((stream._duplexState & DESTROYED) !== 0) ondata(null);
        })
      },
      return () {
        return destroy(null)
      },
      throw (err) {
        return destroy(err)
      }
    }

    function onreadable () {
      if (promiseResolve !== null) ondata(stream.read());
    }

    function onclose () {
      if (promiseResolve !== null) ondata(null);
    }

    function ondata (data) {
      if (promiseReject === null) return
      if (error) promiseReject(error);
      else if (data === null && (stream._duplexState & READ_DONE) === 0) promiseReject(STREAM_DESTROYED);
      else promiseResolve({ value: data, done: data === null });
      promiseReject = promiseResolve = null;
    }

    function destroy (err) {
      stream.destroy(err);
      return new Promise((resolve, reject) => {
        if (stream._duplexState & DESTROYED) return resolve({ value: undefined, done: true })
        stream.once('close', function () {
          if (err) reject(err);
          else resolve({ value: undefined, done: true });
        });
      })
    }
  }
}

class Writable$1 extends Stream {
  constructor (opts) {
    super(opts);

    this._duplexState |= OPENING | READ_DONE;
    this._writableState = new WritableState(this, opts);

    if (opts) {
      if (opts.writev) this._writev = opts.writev;
      if (opts.write) this._write = opts.write;
      if (opts.final) this._final = opts.final;
    }
  }

  _writev (batch, cb) {
    cb(null);
  }

  _write (data, cb) {
    this._writableState.autoBatch(data, cb);
  }

  _final (cb) {
    cb(null);
  }

  static isBackpressured (ws) {
    return (ws._duplexState & WRITE_BACKPRESSURE_STATUS) !== 0
  }

  write (data) {
    this._writableState.updateNextTick();
    return this._writableState.push(data)
  }

  end (data) {
    this._writableState.updateNextTick();
    this._writableState.end(data);
    return this
  }
}

class Duplex extends Readable$1 { // and Writable
  constructor (opts) {
    super(opts);

    this._duplexState = OPENING;
    this._writableState = new WritableState(this, opts);

    if (opts) {
      if (opts.writev) this._writev = opts.writev;
      if (opts.write) this._write = opts.write;
      if (opts.final) this._final = opts.final;
    }
  }

  _writev (batch, cb) {
    cb(null);
  }

  _write (data, cb) {
    this._writableState.autoBatch(data, cb);
  }

  _final (cb) {
    cb(null);
  }

  write (data) {
    this._writableState.updateNextTick();
    return this._writableState.push(data)
  }

  end (data) {
    this._writableState.updateNextTick();
    this._writableState.end(data);
    return this
  }
}

class Transform$1 extends Duplex {
  constructor (opts) {
    super(opts);
    this._transformState = new TransformState(this);

    if (opts) {
      if (opts.transform) this._transform = opts.transform;
      if (opts.flush) this._flush = opts.flush;
    }
  }

  _write (data, cb) {
    if (this._readableState.buffered >= this._readableState.highWaterMark) {
      this._transformState.data = data;
    } else {
      this._transform(data, this._transformState.afterTransform);
    }
  }

  _read (cb) {
    if (this._transformState.data !== null) {
      const data = this._transformState.data;
      this._transformState.data = null;
      cb(null);
      this._transform(data, this._transformState.afterTransform);
    } else {
      cb(null);
    }
  }

  _transform (data, cb) {
    cb(null, data);
  }

  _flush (cb) {
    cb(null);
  }

  _final (cb) {
    this._transformState.afterFinal = cb;
    this._flush(transformAfterFlush.bind(this));
  }
}

class PassThrough extends Transform$1 {}

function transformAfterFlush (err, data) {
  const cb = this._transformState.afterFinal;
  if (err) return cb(err)
  if (data !== null && data !== undefined) this.push(data);
  this.push(null);
  cb(null);
}

function pipelinePromise (...streams) {
  return new Promise((resolve, reject) => {
    return pipeline(...streams, (err) => {
      if (err) return reject(err)
      resolve();
    })
  })
}

function pipeline (stream, ...streams) {
  const all = Array.isArray(stream) ? [...stream, ...streams] : [stream, ...streams];
  const done = (all.length && typeof all[all.length - 1] === 'function') ? all.pop() : null;

  if (all.length < 2) throw new Error('Pipeline requires at least 2 streams')

  let src = all[0];
  let dest = null;
  let error = null;

  for (let i = 1; i < all.length; i++) {
    dest = all[i];

    if (isStreamx(src)) {
      src.pipe(dest, onerror);
    } else {
      errorHandle(src, true, i > 1, onerror);
      src.pipe(dest);
    }

    src = dest;
  }

  if (done) {
    let fin = false;

    dest.on('finish', () => { fin = true; });
    dest.on('error', err => { error = error || err; });
    dest.on('close', () => done(error || (fin ? null : PREMATURE_CLOSE)));
  }

  return dest

  function errorHandle (s, rd, wr, onerror) {
    s.on('error', onerror);
    s.on('close', onclose);

    function onclose () {
      if (rd && s._readableState && !s._readableState.ended) return onerror(PREMATURE_CLOSE)
      if (wr && s._writableState && !s._writableState.ended) return onerror(PREMATURE_CLOSE)
    }
  }

  function onerror (err) {
    if (!err || error) return
    error = err;

    for (const s of all) {
      s.destroy(err);
    }
  }
}

function isStream (stream) {
  return !!stream._readableState || !!stream._writableState
}

function isStreamx (stream) {
  return typeof stream._duplexState === 'number' && isStream(stream)
}

function isReadStreamx (stream) {
  return isStreamx(stream) && stream.readable
}

function isTypedArray (data) {
  return typeof data === 'object' && data !== null && typeof data.byteLength === 'number'
}

function defaultByteLength (data) {
  return isTypedArray(data) ? data.byteLength : 1024
}

function noop$2 () {}

function abort () {
  this.destroy(new Error('Stream aborted.'));
}

var streamx = {
  pipeline,
  pipelinePromise,
  isStream,
  isStreamx,
  Stream,
  Writable: Writable$1,
  Readable: Readable$1,
  Duplex,
  Transform: Transform$1,
  // Export PassThrough for compatibility with Node.js core's stream module
  PassThrough
};

function bindSymbol$3 (obj, symbol, value) {
  Object.defineProperty(obj, symbol, {
    enumerable: false,
    configurable: false,
    writable: false,
    value
  });
}

class MapSet$1 {
  constructor () {
    this._map = new Map();
  }

  get (key) {
    if (this._map.has(key)) return this._map.get(key)
    return new Set()
  }

  add (key, value) {
    if (!this._map.has(key)) this._map.set(key, new Set());
    this.get(key).add(value);
  }

  has (key, value) {
    if (!this._map.has(key)) return false
    return this._map.get(key).has(value)
  }

  highest (key) {
    if (!this._map.has(key)) return null
    return Array.from(this._map.get(key)).sort().pop()
  }
}

var util$1 = { bindSymbol: bindSymbol$3, MapSet: MapSet$1 };

var address = { parseAddress: parseAddress$2, encodeAddress: encodeAddress$2 };
function parseAddress$2 (address) {
  if (!address) throw new Error('Cannot parse empty address')

  // Allow an object with address property
  if (typeof address === 'object' && typeof address.address === 'string') {
    address = address.address;
  // Allow an object with { namespace, type, field, version } properties
  } else if (typeof address === 'object') {
    return parseAddress$2(encodeAddress$2(address))
  }

  // Parse the address format. Example:
  // sonar/Entity@3#label
  const regex = /^(?:([^/#@]+)\/)?([^@#/]+)(?:@(\d+))?(?:#([^/#@]+))?$/;
  const matches = address.match(regex);
  if (!matches) throw new Error('Invalid address')
  matches.shift();
  let [namespace, type, version, field] = matches;
  if (version) version = parseInt(version);
  return { namespace, type, field, version }
}

function encodeAddress$2 (parts) {
  let { namespace, type, field, version } = parts;
  // TODO: Think if we want to accept both name and type
  if (!type && parts.name) type = parts.name;
  if (!type || !validSegment(type)) {
    throw new Error('Cannot encode address: invalid type name')
  }
  if (namespace && !validSegment(namespace)) {
    throw new Error('Cannot encode address: invalid namespace')
  }
  if (field && !validSegment(field)) {
    throw new Error('Cannot encode address: invalid field name')
  }
  if (!version) version = 0;
  let address = '';
  if (namespace) address += namespace + '/';
  if (type) address += type;
  if (version !== undefined) address += '@' + version;
  if (field) address += '#' + field;
  return address
}

function validSegment (segment) {
  return !hasChar(segment, '/') && !hasChar(segment, '#') && !hasChar(segment, '@')
}

function hasChar (str, char) {
  return str.indexOf(char) !== -1
}

const RECORD$1 = Symbol('sonar-record');
const TYPE = Symbol('sonar-type');
const FIELD = Symbol('sonar-field');
const SC$3 = Symbol('sonar-schema');

var symbols = {
  RECORD: RECORD$1,
  TYPE,
  SC: SC$3,
  FIELD
};

const { bindSymbol: bindSymbol$2 } = util$1;
const { parseAddress: parseAddress$1, encodeAddress: encodeAddress$1 } = address;
const { SC: SC$2 } = symbols;

class Type {
  static jsonSchemaToSpec (spec) {
    return jsonSchemaToSpec(spec)
  }

  static isJsonSchema (spec) {
    return isJsonSchema(spec)
  }

  // static fromJSONSchema (schema, spec) {
  //   return new Type(schema, jsonSchemaToSpec(spec))
  // }

  constructor (schema, spec) {
    bindSymbol$2(this, SC$2, schema);
    this._fields = new Set();

    if (spec.address) {
      const parts = parseAddress$1(spec.address);
      this._namespace = parts.namespace;
      this._name = parts.type;
      this._version = parts.version || spec.version || 0;
    } else {
      // This will throw if namespace is undefined and default namespace is not set.
      this._namespace = spec.namespace || this[SC$2].defaultNamespace();
      this._name = spec.name;
      this._version = spec.version || 0;
    }

    if (!this._name) throw new Error('Cannot create Type: missing name')

    this._address = encodeAddress$1({
      namespace: this._namespace,
      type: this._name,
      // TODO: Think through if we want the version in the address.
      version: this._version
    });

    if (spec.fields) {
      for (const [name, fieldSpec] of Object.entries(spec.fields)) {
        fieldSpec.name = name;
        const field = this[SC$2]._addFieldForType(this, fieldSpec);
        this._fields.add(field.address);
      }
    }

    if (spec.refines) {
      this._parent = this[SC$2].resolveTypeAddress(spec.refines);
    }

    this._info = {
      title: spec.title,
      description: spec.description
    };
  }

  get title () {
    return this._info.title || this._name
  }

  get name () {
    return this._name
  }

  get namespace () {
    return this._namespace
  }

  get description () {
    return this._info.description
  }

  get address () {
    return this._address
  }

  get version () {
    return this._version
  }

  parentType () {
    if (!this._parent) return null
    return this[SC$2].getType(this._parent)
  }

  allParents () {
    const addresses = [this.address];
    const parentType = this.parentType();
    if (parentType) addresses.push(...parentType.allParents());
    return addresses
  }

  // TODO: Do the resolution only once, not on each call?
  // TODO: Or: Use an lazy iterator.
  fields () {
    const fields = Array.from(this._fields)
      .map(address => this[SC$2].getField(address));

    const parentType = this.parentType();
    if (parentType) {
      fields.push(...parentType.fields());
    }

    return fields
  }

  fieldAddresses () {
    return this.fields.map(field => field.address)
  }

  toJSONSchema () {
    const spec = {
      $id: this.address,
      properties: {}
    };
    if (this.title) spec.title = this.title;
    if (this.description) spec.description = this.description;
    for (const field of this.fields()) {
      spec.properties[field.name] = field.toJSONSchema();
    }
    return spec
  }

  toJSON () {
    return {
      address: this.address,
      title: this.title,
      description: this.description,
      refines: this._parent,
      fields: this.fields().reduce((all, field) => {
        all[field.name] = field.toJSON();
        return all
      }, {})
    }
  }
}

class MissingType extends Type {
  constructor (schema) {
    super(schema, { name: '_missing', missing: true });
  }
}

Type.MissingType = MissingType;

var type = Type;
var jsonSchemaToSpec_1 = jsonSchemaToSpec;

function jsonSchemaToSpec (spec) {
  if (!spec.fields) spec.fields = {};
  if (spec.properties) {
    for (let [name, fieldSpec] of Object.entries(spec.properties)) {
      if (fieldSpec.sonar) {
        fieldSpec = Object.assign(fieldSpec, fieldSpec.sonar);
        fieldSpec.sonar = undefined;
      }
      spec.fields[name] = fieldSpec;
    }
    spec.properties = undefined;
  }
  if (spec.$id) {
    spec.address = spec.$id;
    spec.$id = undefined;
  }
  if (spec.sonar) {
    spec = Object.assign(spec, spec.sonar);
    spec.sonar = undefined;
  }
  return spec
}

function isJsonSchema (spec) {
  if (spec.properties) return true
  return false
}
type.jsonSchemaToSpec = jsonSchemaToSpec_1;

var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

function createCommonjsModule(fn) {
  var module = { exports: {} };
	return fn(module, module.exports), module.exports;
}

var objectPath = createCommonjsModule(function (module) {
(function (root, factory){

  /*istanbul ignore next:cant test*/
  {
    module.exports = factory();
  }
})(commonjsGlobal, function(){

  var toStr = Object.prototype.toString;
  function hasOwnProperty(obj, prop) {
    if(obj == null) {
      return false
    }
    //to handle objects with null prototypes (too edge case?)
    return Object.prototype.hasOwnProperty.call(obj, prop)
  }

  function isEmpty(value){
    if (!value) {
      return true;
    }
    if (isArray(value) && value.length === 0) {
        return true;
    } else if (typeof value !== 'string') {
        for (var i in value) {
            if (hasOwnProperty(value, i)) {
                return false;
            }
        }
        return true;
    }
    return false;
  }

  function toString(type){
    return toStr.call(type);
  }

  function isObject(obj){
    return typeof obj === 'object' && toString(obj) === "[object Object]";
  }

  var isArray = Array.isArray || function(obj){
    /*istanbul ignore next:cant test*/
    return toStr.call(obj) === '[object Array]';
  };

  function isBoolean(obj){
    return typeof obj === 'boolean' || toString(obj) === '[object Boolean]';
  }

  function getKey(key){
    var intKey = parseInt(key);
    if (intKey.toString() === key) {
      return intKey;
    }
    return key;
  }

  function factory(options) {
    options = options || {};

    var objectPath = function(obj) {
      return Object.keys(objectPath).reduce(function(proxy, prop) {
        if(prop === 'create') {
          return proxy;
        }

        /*istanbul ignore else*/
        if (typeof objectPath[prop] === 'function') {
          proxy[prop] = objectPath[prop].bind(objectPath, obj);
        }

        return proxy;
      }, {});
    };

    var hasShallowProperty;
    if (options.includeInheritedProps) {
      hasShallowProperty = function () {
        return true
      };
    } else {
      hasShallowProperty = function (obj, prop) {
        return (typeof prop === 'number' && Array.isArray(obj)) || hasOwnProperty(obj, prop)
      };
    }

    function getShallowProperty(obj, prop) {
      if (hasShallowProperty(obj, prop)) {
        return obj[prop];
      }
    }

    function set(obj, path, value, doNotReplace){
      if (typeof path === 'number') {
        path = [path];
      }
      if (!path || path.length === 0) {
        return obj;
      }
      if (typeof path === 'string') {
        return set(obj, path.split('.').map(getKey), value, doNotReplace);
      }
      var currentPath = path[0];
      var currentValue = getShallowProperty(obj, currentPath);
      if (options.includeInheritedProps && (currentPath === '__proto__' ||
        (currentPath === 'constructor' && typeof currentValue === 'function'))) {
        throw new Error('For security reasons, object\'s magic properties cannot be set')
      }
      if (path.length === 1) {
        if (currentValue === void 0 || !doNotReplace) {
          obj[currentPath] = value;
        }
        return currentValue;
      }

      if (currentValue === void 0) {
        //check if we assume an array
        if(typeof path[1] === 'number') {
          obj[currentPath] = [];
        } else {
          obj[currentPath] = {};
        }
      }

      return set(obj[currentPath], path.slice(1), value, doNotReplace);
    }

    objectPath.has = function (obj, path) {
      if (typeof path === 'number') {
        path = [path];
      } else if (typeof path === 'string') {
        path = path.split('.');
      }

      if (!path || path.length === 0) {
        return !!obj;
      }

      for (var i = 0; i < path.length; i++) {
        var j = getKey(path[i]);

        if((typeof j === 'number' && isArray(obj) && j < obj.length) ||
          (options.includeInheritedProps ? (j in Object(obj)) : hasOwnProperty(obj, j))) {
          obj = obj[j];
        } else {
          return false;
        }
      }

      return true;
    };

    objectPath.ensureExists = function (obj, path, value){
      return set(obj, path, value, true);
    };

    objectPath.set = function (obj, path, value, doNotReplace){
      return set(obj, path, value, doNotReplace);
    };

    objectPath.insert = function (obj, path, value, at){
      var arr = objectPath.get(obj, path);
      at = ~~at;
      if (!isArray(arr)) {
        arr = [];
        objectPath.set(obj, path, arr);
      }
      arr.splice(at, 0, value);
    };

    objectPath.empty = function(obj, path) {
      if (isEmpty(path)) {
        return void 0;
      }
      if (obj == null) {
        return void 0;
      }

      var value, i;
      if (!(value = objectPath.get(obj, path))) {
        return void 0;
      }

      if (typeof value === 'string') {
        return objectPath.set(obj, path, '');
      } else if (isBoolean(value)) {
        return objectPath.set(obj, path, false);
      } else if (typeof value === 'number') {
        return objectPath.set(obj, path, 0);
      } else if (isArray(value)) {
        value.length = 0;
      } else if (isObject(value)) {
        for (i in value) {
          if (hasShallowProperty(value, i)) {
            delete value[i];
          }
        }
      } else {
        return objectPath.set(obj, path, null);
      }
    };

    objectPath.push = function (obj, path /*, values */){
      var arr = objectPath.get(obj, path);
      if (!isArray(arr)) {
        arr = [];
        objectPath.set(obj, path, arr);
      }

      arr.push.apply(arr, Array.prototype.slice.call(arguments, 2));
    };

    objectPath.coalesce = function (obj, paths, defaultValue) {
      var value;

      for (var i = 0, len = paths.length; i < len; i++) {
        if ((value = objectPath.get(obj, paths[i])) !== void 0) {
          return value;
        }
      }

      return defaultValue;
    };

    objectPath.get = function (obj, path, defaultValue){
      if (typeof path === 'number') {
        path = [path];
      }
      if (!path || path.length === 0) {
        return obj;
      }
      if (obj == null) {
        return defaultValue;
      }
      if (typeof path === 'string') {
        return objectPath.get(obj, path.split('.'), defaultValue);
      }

      var currentPath = getKey(path[0]);
      var nextObj = getShallowProperty(obj, currentPath);
      if (nextObj === void 0) {
        return defaultValue;
      }

      if (path.length === 1) {
        return nextObj;
      }

      return objectPath.get(obj[currentPath], path.slice(1), defaultValue);
    };

    objectPath.del = function del(obj, path) {
      if (typeof path === 'number') {
        path = [path];
      }

      if (obj == null) {
        return obj;
      }

      if (isEmpty(path)) {
        return obj;
      }
      if(typeof path === 'string') {
        return objectPath.del(obj, path.split('.'));
      }

      var currentPath = getKey(path[0]);
      if (!hasShallowProperty(obj, currentPath)) {
        return obj;
      }

      if(path.length === 1) {
        if (isArray(obj)) {
          obj.splice(currentPath, 1);
        } else {
          delete obj[currentPath];
        }
      } else {
        return objectPath.del(obj[currentPath], path.slice(1));
      }

      return obj;
    };

    return objectPath;
  }

  var mod = factory();
  mod.create = factory;
  mod.withInheritedProps = factory({includeInheritedProps: true});
  return mod;
});
});

const { bindSymbol: bindSymbol$1 } = util$1;
const { SC: SC$1 } = symbols;

class Field {
  constructor (schema, spec) {
    if (!spec.address) throw new Error('Field address is required')
    if (!spec.name) throw new Error('Field name is required')
    if (!spec.defaultWidget) {
      spec.defaultWidget = 'TextWidget';
    }
    bindSymbol$1(this, SC$1, schema);
    this._address = spec.address;
    this._defaultWidget = spec.defaultWidget;
    this._spec = spec;
    this._children = new Set();
    this._parent = null;
  }

  get name () {
    return this._spec.name
  }

  get fieldType () {
    return this._spec.type
  }

  get title () {
    return this._spec.title || this._spec.name
  }

  get description () {
    return this._spec.description
  }

  get address () {
    return this._address
  }

  get defaultWidget () {
    return this._defaultWidget
  }

  get index () {
    return this._spec.index || {}
  }

  allVariants () {
    const addresses = [this.address];
    addresses.push(...this._children);
    if (this._parent) addresses.push(this[SC$1].getField(this._parent).allVariants());
    return addresses
  }

  getParent () {
    if (!this._parent) return null
    return this[SC$1].getField(this._parent)
  }

  _build (strict = true, force = false) {
    if (!force && this._built) return
    this._mergeRefines(strict);
    this._built = true;
  }

  _mergeRefines (strict = true) {
    // Nothing to do for fields without parents.
    if (!this._spec.refines) return

    // TODO: This will endlessly loop for nested refineds.
    // It should throw or stop when recursion is encountered.
    const parent = this[SC$1].getField(this._spec.refines);

    if (strict && !parent) {
      throw new Error(`Missing parent field ${this._spec.refines} while resolving ${this.address}`)
    } else if (!parent) return

    parent._build();
    this._parent = parent.address;
    parent._children.add(this.address);

    // for (const [key, value] of Object.entries(parent)) {
    //   if (this._spec[key] === undefined) {
    //     this._spec[key] = value
    //   }
    // }
  }

  getType () {
    const typeAddress = this._address.split('#')[0];
    return this[SC$1].getType(typeAddress)
  }

  setProp (path, value) {
    objectPath.set(this._spec, path, value);
  }

  getProp (path) {
    objectPath.get(this._spec);
  }

  toJSONSchema () {
    const spec = {};
    spec.title = this.title;
    spec.description = this.description;

    // TODO: Define list of properties allowed.
    const sonarInfo = {
      refines: this._parent,
      index: this._spec.index
    };

    // TODO: Likely we want to have this handled by field type classes,
    // so this.fieldType.jsonType()
    if (this.fieldType === 'relation') {
      spec.type = 'string';
      sonarInfo.fieldType = this.fieldType;
    } else {
      spec.type = this.fieldType;
    }
    spec.sonar = sonarInfo;
    return spec
  }

  toJSON () {
    return this._spec
  }

  // static fromJSONSchema (schema, json) {
  //   const spec = {
  //     ...json,
  //     fields: {},
  //     ...json.sonar || {}
  //   }
  //   for (const [name, fieldSpec] of Object.entries(spec)) {
  //     spec.fields[name] = {
  //       ...fieldSpec,
  //       ...fieldSpec.sonar || {}
  //     }
  //     if (!fieldSpec.fieldType && fieldSpec.type) {
  //       fieldSpec.fieldType = fieldSpec.type
  //     }
  //   }
  //   spec.properties = undefined
  //   return new Field(schema, spec)
  // }
}

var field = Field;

var prettyHash = function prettyHash (buf) {
  if (Buffer.isBuffer(buf)) buf = buf.toString('hex');
  if (typeof buf === 'string' && buf.length > 8) {
    return buf.slice(0, 6) + '..' + buf.slice(-2)
  }
  return buf
};

var browser$2 = Symbol.for('nodejs.util.inspect.custom');

const kSubs = Symbol('kSubs');

var emitter = class SimpleEmitter {
  constructor () {
    this[kSubs] = new Set();
  }

  subscribe (fn, triggerNow = false) {
    this[kSubs].add(fn);
    if (triggerNow) fn(this);
    return () => (this[kSubs].delete(fn))
  }

  emit () {
    for (const fn of this[kSubs]) {
      fn(this);
    }
  }
};

var versions = class Versions extends emitter {
  constructor (opts = {}) {
    super();
    this._versions = new Map();
    this._current = new Set();
    this._outdated = new Set();
  }

  current () {
    return Array.from([...this._current.values()])
  }

  all () {
    return Array.from([...this._versions.values()])
  }

  latest () {
    return this.current().reduce((latest, current) => {
      if (!latest) return current
      if (current.timestamp > latest.timestamp) return current
      return latest
    })
  }

  has (version) {
    if (this._versions.has(version.address)) return true
    return false
  }

  put (version) {
    if (this._versions.has(version.address)) return
    this._versions.set(version.address, version);
    let isCurrent = true;
    if (!version.links) version.links = [];
    if (this._outdated.has(version.address)) {
      isCurrent = false;
    } else {
      for (const current of this._current) {
        if (current.links.indexOf(version.address) !== -1) {
          isCurrent = false;
        }
      }
    }
    for (const link of version.links) {
      this._outdated.add(link);
    }
    if (isCurrent) {
      for (const link of version.links) {
        if (this._versions.has(link)) {
          this._current.delete(this._versions.get(link));
        }
      }
      this._current.add(version);
    }
    this.emit();
  }
};

const { SC, RECORD } = symbols;


const { bindSymbol } = util$1;



function fmtAddress (address) {
  const [key, seq] = address.split('@');
  return key.substring(0, 5) + '..' + key.substring(30, 32) + '@' + seq
}

// Base class for Record and Entity
class Node extends emitter {
  constructor (schema) {
    super();
    bindSymbol(this, SC, schema);
  }

  field (name, single = true) {
    return this._field(name, single)
  }

  fields (name) {
    return this._field(name, false)
  }

  hasField (name) {
    try {
      return this.fields(name).length > 0
    } catch (err) {
      return false
    }
  }

  values (name) {
    return this.fields(name).values()
  }

  mapFields (name, fn) {
    if (!fn) {
      fn = name;
      name = null;
    }
    return this.fields(name).map(fn)
  }

  // TODO: Depcreate in favor of getOne/getMany.
  get (name, single = true) {
    return this.field(name, single).value
  }

  getOne (name) {
    return this.get(name, true)
  }

  getMany (name) {
    return this.get(name, false)
  }

  gotoOne (store, name) {
    return this._goto(store, name, true)
  }

  gotoMany (store, name) {
    return this._goto(store, name, false)
  }

  // Invoked by the Node high-level methods.
  _goto (store, name, single = false) {
    const field = this.field(name);
    // TODO: This means crash.
    if (!field) throw new Error('Field not found: ' + name)
    if (!field.type === 'relation') {
      throw new Error('Not a relation field: ' + name)
    }
    // TODO: Deal with multiple values
    const targetAddresses = field.values();
    const targetEntities = targetAddresses.map(address => {
      let entity = store.getEntity(address);
      if (!entity) entity = new MissingEntity(this[SC], address);
      return entity
    });
    // TODO: targetEntities are Entity not FieldValue objects,
    // not sure if that is good or if it should be a Proxy or so
    // on FieldValue.
    const fieldValues = new FieldValueList(this[SC], ...targetEntities);
    if (single) return fieldValues.first()
    else return fieldValues.filter(f => !f.empty())
  }
}

class FieldValueList extends Array {
  constructor (schema, ...values) {
    super(...values);
    bindSymbol(this, SC, schema);
  }

  get value () {
    return this.values()
  }

  values () {
    return this.map(fv => fv.value)
  }

  first () {
    if (!this.length) return new MissingFieldValue(this._field)
    // TODO: How do we find the best value?
    return this[0]
  }
}

class Record$3 extends Node {
  constructor (schema, initialVersion) {
    super(schema);
    this._versions = new versions();
    this.addVersion(initialVersion);
    this._versions.subscribe(() => this.emit());
  }

  versions () {
    return this._versions.current()
  }

  allVersions () {
    return this._versions.all()
  }

  get latest () {
    return this._latest
  }

  addVersion (recordVersion) {
    if (!(recordVersion instanceof RecordVersion$3)) {
      recordVersion = new RecordVersion$3(this[SC], recordVersion);
    }
    if (!this._id) {
      this._id = recordVersion.id;
      this._type = recordVersion.type;
    }
    if (recordVersion.path !== this.path) {
      throw new Error('RecordVersion does not match Record path')
    }
    this._versions.put(recordVersion);
  }

  hasVersion (address) {
    this._versions.has(address);
  }

  get path () {
    return this._type + '/' + this._id
  }

  get _latest () {
    return this._versions.latest()
  }

  get id () {
    return this._latest.id
  }

  get value () {
    return this._latest.value
  }

  set value (value) {
    this._latest.value = value;
  }

  get type () {
    return this._latest.type
  }

  get address () {
    return this._latest.adress
  }

  get shortAddress () {
    return this._latest.shortAddress
  }

  get deleted () {
    return this._latest.deleted
  }

  get key () {
    return this._latest.key
  }

  set key (key) {
    this._latest.key = key;
  }

  get feed () {
    return this._latest.key
  }

  get seq () {
    return this._latest.seq
  }

  get lseq () {
    return this._latest.lseq
  }

  get timestamp () {
    return this._latest.timestamp
  }

  set timestamp (timestamp) {
    this._latest.timestamp = timestamp;
  }

  get meta () {
    return this._meta || this._latest.meta || {}
  }

  set lseq (lseq) {
    this._latest.lseq = lseq;
  }

  get links () {
    return this._latest.links
  }

  set links (links) {
    this._latest.links = links;
  }

  getType () {
    return this._latest.getType()
  }

  hasType (typeAddress) {
    return this._latest.hasType(typeAddress)
  }

  // Invoked by the Node high-level methods.
  _field (fieldName, single = true) {
    return this._latest._field(fieldName, single)
  }

  toJSON () {
    return this._latest.toJSON()
  }
}

class RecordVersion$3 extends Node {
  constructor (schema, record) {
    if (record instanceof RecordVersion$3) record = record._record;

    if (!record.type) {
      throw new Error('Cannot upcast record: Missing type')
    }
    if (!record.id) {
      throw new Error('Cannot upcast record: Missing id')
    }

    // TODO: Decide how to deal with unknown types - likely we want to error here.
    // Maybe we also want to support deriving types on first use? Or allow to deal
    // with records with an unknown type in a limited fashion.
    if (!schema.getType(record.type)) {
      throw new Error(`Cannot upcast record: Unknown type "${record.type}"`)
    }
    record.type = schema.resolveTypeAddress(record.type);

    super(schema);

    this._record = record;
    this._built = false;

    // Prevent double-upcasting
    // TODO: Find out if this is a performance concern.
    this[RECORD] = this;
    record[RECORD] = this;
  }

  update (nextValue) {
    nextValue = Object.assign({}, this._value, nextValue);
    const nextVersion = {
      type: this.type,
      id: this.id,
      value: nextValue
    };
    return new RecordVersion$3(this[SC], nextVersion)
  }

  get id () {
    return this._record.id
  }

  get value () {
    return this._record.value
  }

  set value (value) {
    this._record.value = value;
  }

  get type () {
    return this._record.type
  }

  get address () {
    return this._record.key + '@' + this._record.seq
  }

  get shortAddress () {
    return fmtAddress(this.address)
  }

  get path () {
    return this.type + '/' + this.id
  }

  get deleted () {
    return this._record.deleted
  }

  get key () {
    return this._record.key
  }

  set key (key) {
    this._record.key = key;
  }

  get feed () {
    return this._record.key
  }

  get seq () {
    return this._record.seq
  }

  get lseq () {
    return this._record.lseq
  }

  get timestamp () {
    return this._record.timestamp
  }

  set timestamp (timestamp) {
    this._record.timestamp = timestamp;
  }

  get meta () {
    return this._meta || this._record.meta || {}
  }

  set meta (obj) {
    if (!this._meta) this._meta = this.meta;
    this._meta = { ...this.meta, ...obj };
  }

  // TODO: kappa-scopes/scopes.js:518 calls this to add lseq from indexer state.
  // Review how adding this info should work.
  set lseq (lseq) {
    this._record.lseq = lseq;
  }

  get links () {
    return this._record.links
  }

  set links (links) {
    this._record.links = links;
  }

  getType () {
    return this[SC].getType(this.type)
  }

  hasType (typeAddress) {
    typeAddress = this[SC].resolveTypeAddress(typeAddress);
    if (this.type === typeAddress) return true
    const type = this.getType();
    if (!type) return false
    const allTypes = type.allParents();
    return allTypes.indexOf(typeAddress) !== -1
  }

  allTypes () {
    const type = this.getType();
    if (!type) return false
    const allTypes = type.allParents();
    return allTypes
  }

  // Invoked by the Node high-level methods.
  _field (fieldName, single = true) {
    const fieldValues = this._filterFieldValues(fieldName);
    if (single) return fieldValues.first()
    return fieldValues
  }

  _update (force = false) {
    if (!force && this._built) return
    this._buildFieldValues();
    this._built = true;
  }

  _buildFieldValues () {
    const fields = this.getType().fields();
    this._fieldValues = new FieldValueList(this[SC]);
    if (!this._record.value || typeof this._record.value !== 'object') {
      return
    }
    for (const field of fields) {
      field._build(false);
      if (this._record.value[field.name] !== undefined) {
        // const fieldValue = new FieldValue(field, this._record.value[field.name])
        const fieldValue = new FieldValue(field, this._record.value[field.name], this);
        this._fieldValues.push(fieldValue);
      }
    }
  }

  _findField (name) {
    if (name.indexOf('#') !== -1) {
      return this[SC].getField(name)
    }
    for (const field of this.getType().fields()) {
      if (field.name === name) return field
    }
    return null
  }

  _filterFieldValues (fieldName) {
    this._update();
    if (!fieldName) return this._fieldValues

    const field = this._findField(fieldName);
    if (!field) return new FieldValueList(this[SC])

    const validAddresses = field.allVariants();
    return this._fieldValues.filter(field => {
      return validAddresses.indexOf(field.fieldAddress) !== -1
    })
  }

  toJSON () {
    // TODO: Add opts to skip encoding lseq on put.
    // TODO: We don't need both address and key, seq.
    return {
      // stored keys
      id: this.id,
      type: this.type,
      value: this.value,
      links: this.links,
      deleted: this.deleted,
      timestamp: this.timestamp,

      // wire key
      key: this.key,
      seq: this.seq,
      lseq: this.lseq,
      meta: this.meta
    }
  }

  [browser$2] (depth, opts = {}) {
    if (!opts.stylize) opts.stylize = obj => obj;
    const { stylize } = opts;
    let ind = '';
    if (typeof opts.indentationLvl === 'number') {
      while (ind.length < opts.indentationLvl) ind += ' ';
    }
    const h = str => stylize(str, 'special');
    const s = str => stylize(str);
    const links = this.links ? this.links.length : 0;
    const value = this.deleted ? '<deleted>' : JSON.stringify(this.value).substring(0, 320);
    const meta = s('feed ') + h(prettyHash(this.key)) + s('@') + this.seq +
      s(' lseq ') + (this.lseq || '--') +
      s(' links ') + links;
    return `RecordVersion(
${ind}  ${s('type')} ${this.type} ${s('id')} ${this.id}
${ind}  ${s('value')} ${value}
${ind}  ${meta}
)`
  }
}

class FieldValue {
  constructor (field, value, record) {
    this._field = field;
    this._value = value;
    this._record = record;
  }

  get value () {
    // return this._record.value[this._field.name]
    return this._value
  }

  values () {
    const value = this.value;
    return Array.isArray(value) ? value : [value]
  }

  get field () {
    return this._field
  }

  get fieldAddress () {
    return this._field.address
  }

  get address () {
    return this._field.address
  }

  get fieldType () {
    return this._field.fieldType
  }

  get defaultWidget () {
    return this._field.defaultWidget
  }

  get name () {
    return this._field.name
  }

  get title () {
    return this._field.title || this._field.name
  }

  get record () {
    return this._record
  }

  empty () {
    return this._field.missing || this._value === undefined
  }

  first () {
    return this
  }
}

class MissingFieldValue extends FieldValue {
  constructor (field = {}) {
    field = Object.assign({
      title: '(missing field)',
      name: '_missing'
    }, field, { missing: true });
    super(field, undefined);
  }
}

class Entity$2 extends Node {
  constructor (schema, records) {
    super(schema);
    this._records = new Set();
    this._id = null;
    if (records) {
      for (const record of records) {
        this.add(record);
      }
    }
  }

  get id () {
    return this._id
  }

  get address () {
    return this._id
  }

  add (record) {
    if (!this._id) {
      this._id = record.id;
    } else if (this._id !== record.id) {
      throw new Error('Cannot add record to entity: IDs do not match')
    }
    this._records.add(record);
  }

  empty () {
    return !this._missing && !this._records.size
  }

  hasType (typeAddress) {
    for (const record of this._records) {
      if (record.hasType(typeAddress)) return true
    }
    return false
  }

  getTypes () {
    return Array.from(new Set(
      Array.from(this._records).map(r => r.getType())
    ))
  }

  // Invoked by the Node high-level methods.
  _field (fieldName, single = true) {
    const fieldValues = new FieldValueList(this[SC]);

    for (const record of this._records) {
      if (fieldName && record.hasField(fieldName)) {
        fieldValues.push(...record.field(fieldName, false));
      } else if (!fieldName) {
        fieldValues.push(...record.fields());
      }
    }

    if (single) return fieldValues.first()
    return fieldValues
  }
}

class MissingEntity extends Entity$2 {
  constructor (schema, id) {
    super(schema);
    this._missing = true;
    this._id = id;
  }
}

var records = {
  Entity: Entity$2,
  MissingEntity,
  Record: Record$3,
  RecordVersion: RecordVersion$3,
  FieldValue,
  FieldValueList,
  MissingFieldValue
};

const { MapSet } = util$1;
const { parseAddress, encodeAddress } = address;

// const Store = require('./store')


const { Record: Record$2, RecordVersion: RecordVersion$2, Entity: Entity$1 } = records;

var schema = class Schema {
  constructor (opts = {}) {
    this._types = new Map();
    this._fields = new Map();
    this._typeVersions = new MapSet();
    this._defaultNamespace = opts.defaultNamespace;
    this._onchange = opts.onchange || noop$1;
  }

  Record (record) {
    return new Record$2(this, record)
  }

  RecordVersion (recordVersion) {
    return new RecordVersion$2(this, recordVersion)
  }

  Entity (records) {
    return new Entity$1(this, records)
  }

  setDefaultNamespace (namespace) {
    this._defaultNamespace = namespace;
  }

  defaultNamespace () {
    if (!this._defaultNamespace) throw new Error('Default namespace is not set')
    return this._defaultNamespace
  }

  _parseAddress (address) {
    const parts = parseAddress(address);
    if (!parts.namespace) parts.namespace = this._defaultNamespace;
    if (parts.version === undefined) {
      parts.version = this._typeVersions.highest(parts.namespace + '/' + parts.type);
    }
    return parts
  }

  resolveTypeAddress (address) {
    const parts = this._parseAddress(address);
    if (parts.field) throw new Error('Not a type address: ' + address)
    return encodeAddress(parts)
  }

  resolveFieldAddress (address) {
    const parts = this._parseAddress(address);
    if (!parts.field) throw new Error('Not a field address: ' + address)
    return encodeAddress(parts)
  }

  addType (spec, opts = {}) {
    if (spec.address && this._types.has(spec.address)) {
      return this._types.get(spec.address)
    }
    // TODO: Remove
    if (type.isJsonSchema(spec)) {
      spec = type.jsonSchemaToSpec(spec);
    }

    const type$1 = new type(this, spec);

    // TODO: Make sure types are immutable.
    this._types.set(type$1.address, type$1);
    this._typeVersions.add(type$1.namespace + '/' + type$1.name, type$1.version);
    if (opts.onchange !== false) this._onchange(this);
    return type$1
  }

  // TODO: Remove
  // _addTypeFromJsonSchema (spec) {
  //   spec = Type.jsonSchemaToSpec(spec)
  //   return this.addType(spec)
  // }

  getType (address) {
    return this._types.get(address) || this._types.get(this.resolveTypeAddress(address))
  }

  hasType (address) {
    return this._types.has(address) || this._types.has(this.resolveTypeAddress(address))
  }

  getTypes () {
    return Array.from(this._types.values())
  }

  // This is called by the Type constructor.
  _addFieldForType (type$1, spec) {
    if (!(type$1 instanceof type)) throw new Error('Cannot add field: invalid type argument')
    if (!spec.name) throw new Error('Cannot add field: name is missing')

    const address = type$1.address + '#' + spec.name;
    if (this.hasField(address)) {
      return this._fields.get(address)
      // throw new Error('Field exists: ' + address)
    }
    spec.address = address;

    const field$1 = new field(this, spec);
    this._fields.set(field$1.address, field$1);
    return field$1
  }

  hasField (address) {
    try {
      return this._fields.has(address) || this._fields.has(this.resolveFieldAddress(address))
    } catch (err) {
      return false
    }
  }

  getField (address) {
    return this._fields.get(address) || this._fields.get(this.resolveFieldAddress(address))
  }

  build (strict = true) {
    for (const field of this._fields.values()) {
      field._build(strict);
    }
  }

  toJSON () {
    const spec = {};
    for (const type of this._types.values()) {
      spec[type.address] = type.toJSON();
    }
    return spec
  }

  addTypes (spec) {
    const types = Array.isArray(spec) ? spec : Object.values(spec);
    for (const type of types) {
      this.addType(type, { onchange: false });
    }
    this._onchange(this);
  }

  // [inspect] (depth, opts) {
  //   const { stylize } = opts
  //   var indent = ''
  //   if (typeof opts.indentationLvl === 'number') {
  //     while (indent.length < opts.indentationLvl) indent += ' '
  //   }

  //   const types = this.getTypes().map(type => {
  //     const fields = type.fields().map(field => `${field.name} (${field.fieldType})`)
  //       .join(', ')
  //     return indent + `${type.address}:\n` + indent + '    ' + fields.substring(0, fields.length - 2)
  //   }).join('\n')

  //   return 'Schema(\n' +
  //         indent + '  defaultNamespace: ' + stylize(this._defaultNamespace, 'string') + '\n' +
  //         types + ')'
  // }
};

function noop$1 () {}

const { Entity, Record: Record$1, RecordVersion: RecordVersion$1 } = records;
var store = class Store {
  constructor (opts = {}) {
    this._opts = opts;
    this._schema = opts.schema;
    this._records = new Map();
    this._entities = new Map();
    this._versions = new Map();
  }

  records () {
    return Array.from(this._records.values())
  }

  entities () {
    return Array.from(this._entities.values())
  }

  getEntity (id) {
    return this._entities.get(id)
  }

  getRecord (path) {
    return this._records.get(path)
  }

  getRecordVersion (address) {
    return this._versions.get(address)
  }

  cacheRecord (recordVersion) {
    // TODO: Rethink if we want this.
    if (recordVersion instanceof Record$1) {
      for (const version of recordVersion.allVersions()) {
        this.cacheRecord(version);
      }
      return this.getRecord(recordVersion.path)
    }

    if (!(recordVersion instanceof RecordVersion$1)) {
      recordVersion = new RecordVersion$1(this._schema, recordVersion);
    }

    if (this._records.has(recordVersion.path)) {
      this._records.get(recordVersion.path).addVersion(recordVersion);
    } else {
      const record = new Record$1(this._schema, recordVersion);
      this._records.set(record.path, record);

      if (!this._entities.has(record.id)) {
        const entity = new Entity(this._schema, [record]);
        this._entities.set(entity.id, entity);
      } else {
        const entity = this._entities.get(record.id);
        entity.add(record);
      }
    }

    const record = this._records.get(recordVersion.path);
    this._versions.set(recordVersion.address, recordVersion);
    return record
  }
};

function tryStringify (o) {
  try { return JSON.stringify(o) } catch(e) { return '"[Circular]"' }
}

var quickFormatUnescaped = format;

function format(f, args, opts) {
  var ss = (opts && opts.stringify) || tryStringify;
  var offset = 1;
  if (typeof f === 'object' && f !== null) {
    var len = args.length + offset;
    if (len === 1) return f
    var objects = new Array(len);
    objects[0] = ss(f);
    for (var index = 1; index < len; index++) {
      objects[index] = ss(args[index]);
    }
    return objects.join(' ')
  }
  if (typeof f !== 'string') {
    return f
  }
  var argLen = args.length;
  if (argLen === 0) return f
  var str = '';
  var a = 1 - offset;
  var lastPos = -1;
  var flen = (f && f.length) || 0;
  for (var i = 0; i < flen;) {
    if (f.charCodeAt(i) === 37 && i + 1 < flen) {
      lastPos = lastPos > -1 ? lastPos : 0;
      switch (f.charCodeAt(i + 1)) {
        case 100: // 'd'
          if (a >= argLen)
            break
          if (lastPos < i)
            str += f.slice(lastPos, i);
          if (args[a] == null)  break
          str += Number(args[a]);
          lastPos = i = i + 2;
          break
        case 79: // 'O'
        case 111: // 'o'
        case 106: // 'j'
          if (a >= argLen)
            break
          if (lastPos < i)
            str += f.slice(lastPos, i);
          if (args[a] === undefined) break
          var type = typeof args[a];
          if (type === 'string') {
            str += '\'' + args[a] + '\'';
            lastPos = i + 2;
            i++;
            break
          }
          if (type === 'function') {
            str += args[a].name || '<anonymous>';
            lastPos = i + 2;
            i++;
            break
          }
          str += ss(args[a]);
          lastPos = i + 2;
          i++;
          break
        case 115: // 's'
          if (a >= argLen)
            break
          if (lastPos < i)
            str += f.slice(lastPos, i);
          str += String(args[a]);
          lastPos = i + 2;
          i++;
          break
        case 37: // '%'
          if (lastPos < i)
            str += f.slice(lastPos, i);
          str += '%';
          lastPos = i + 2;
          i++;
          break
      }
      ++a;
    }
    ++i;
  }
  if (lastPos === -1)
    return f
  else if (lastPos < flen) {
    str += f.slice(lastPos);
  }

  return str
}

var browser$1 = pino;

const _console = pfGlobalThisOrFallback().console || {};
const stdSerializers = {
  mapHttpRequest: mock,
  mapHttpResponse: mock,
  wrapRequestSerializer: passthrough,
  wrapResponseSerializer: passthrough,
  wrapErrorSerializer: passthrough,
  req: mock,
  res: mock,
  err: asErrValue
};

function shouldSerialize (serialize, serializers) {
  if (Array.isArray(serialize)) {
    const hasToFilter = serialize.filter(function (k) {
      return k !== '!stdSerializers.err'
    });
    return hasToFilter
  } else if (serialize === true) {
    return Object.keys(serializers)
  }

  return false
}

function pino (opts) {
  opts = opts || {};
  opts.browser = opts.browser || {};

  const transmit = opts.browser.transmit;
  if (transmit && typeof transmit.send !== 'function') { throw Error('pino: transmit option must have a send function') }

  const proto = opts.browser.write || _console;
  if (opts.browser.write) opts.browser.asObject = true;
  const serializers = opts.serializers || {};
  const serialize = shouldSerialize(opts.browser.serialize, serializers);
  let stdErrSerialize = opts.browser.serialize;

  if (
    Array.isArray(opts.browser.serialize) &&
    opts.browser.serialize.indexOf('!stdSerializers.err') > -1
  ) stdErrSerialize = false;

  const levels = ['error', 'fatal', 'warn', 'info', 'debug', 'trace'];

  if (typeof proto === 'function') {
    proto.error = proto.fatal = proto.warn =
    proto.info = proto.debug = proto.trace = proto;
  }
  if (opts.enabled === false) opts.level = 'silent';
  const level = opts.level || 'info';
  const logger = Object.create(proto);
  if (!logger.log) logger.log = noop;

  Object.defineProperty(logger, 'levelVal', {
    get: getLevelVal
  });
  Object.defineProperty(logger, 'level', {
    get: getLevel,
    set: setLevel
  });

  const setOpts = {
    transmit,
    serialize,
    asObject: opts.browser.asObject,
    levels,
    timestamp: getTimeFunction(opts)
  };
  logger.levels = pino.levels;
  logger.level = level;

  logger.setMaxListeners = logger.getMaxListeners =
  logger.emit = logger.addListener = logger.on =
  logger.prependListener = logger.once =
  logger.prependOnceListener = logger.removeListener =
  logger.removeAllListeners = logger.listeners =
  logger.listenerCount = logger.eventNames =
  logger.write = logger.flush = noop;
  logger.serializers = serializers;
  logger._serialize = serialize;
  logger._stdErrSerialize = stdErrSerialize;
  logger.child = child;

  if (transmit) logger._logEvent = createLogEventShape();

  function getLevelVal () {
    return this.level === 'silent'
      ? Infinity
      : this.levels.values[this.level]
  }

  function getLevel () {
    return this._level
  }
  function setLevel (level) {
    if (level !== 'silent' && !this.levels.values[level]) {
      throw Error('unknown level ' + level)
    }
    this._level = level;

    set(setOpts, logger, 'error', 'log'); // <-- must stay first
    set(setOpts, logger, 'fatal', 'error');
    set(setOpts, logger, 'warn', 'error');
    set(setOpts, logger, 'info', 'log');
    set(setOpts, logger, 'debug', 'log');
    set(setOpts, logger, 'trace', 'log');
  }

  function child (bindings) {
    if (!bindings) {
      throw new Error('missing bindings for child Pino')
    }
    const bindingsSerializers = bindings.serializers;
    if (serialize && bindingsSerializers) {
      var childSerializers = Object.assign({}, serializers, bindingsSerializers);
      var childSerialize = opts.browser.serialize === true
        ? Object.keys(childSerializers)
        : serialize;
      delete bindings.serializers;
      applySerializers([bindings], childSerialize, childSerializers, this._stdErrSerialize);
    }
    function Child (parent) {
      this._childLevel = (parent._childLevel | 0) + 1;
      this.error = bind(parent, bindings, 'error');
      this.fatal = bind(parent, bindings, 'fatal');
      this.warn = bind(parent, bindings, 'warn');
      this.info = bind(parent, bindings, 'info');
      this.debug = bind(parent, bindings, 'debug');
      this.trace = bind(parent, bindings, 'trace');
      if (childSerializers) {
        this.serializers = childSerializers;
        this._serialize = childSerialize;
      }
      if (transmit) {
        this._logEvent = createLogEventShape(
          [].concat(parent._logEvent.bindings, bindings)
        );
      }
    }
    Child.prototype = this;
    return new Child(this)
  }
  return logger
}

pino.levels = {
  values: {
    fatal: 60,
    error: 50,
    warn: 40,
    info: 30,
    debug: 20,
    trace: 10
  },
  labels: {
    10: 'trace',
    20: 'debug',
    30: 'info',
    40: 'warn',
    50: 'error',
    60: 'fatal'
  }
};

pino.stdSerializers = stdSerializers;
pino.stdTimeFunctions = Object.assign({}, { nullTime, epochTime, unixTime, isoTime });

function set (opts, logger, level, fallback) {
  const proto = Object.getPrototypeOf(logger);
  logger[level] = logger.levelVal > logger.levels.values[level]
    ? noop
    : (proto[level] ? proto[level] : (_console[level] || _console[fallback] || noop));

  wrap(opts, logger, level);
}

function wrap (opts, logger, level) {
  if (!opts.transmit && logger[level] === noop) return

  logger[level] = (function (write) {
    return function LOG () {
      const ts = opts.timestamp();
      const args = new Array(arguments.length);
      const proto = (Object.getPrototypeOf && Object.getPrototypeOf(this) === _console) ? _console : this;
      for (var i = 0; i < args.length; i++) args[i] = arguments[i];

      if (opts.serialize && !opts.asObject) {
        applySerializers(args, this._serialize, this.serializers, this._stdErrSerialize);
      }
      if (opts.asObject) write.call(proto, asObject(this, level, args, ts));
      else write.apply(proto, args);

      if (opts.transmit) {
        const transmitLevel = opts.transmit.level || logger.level;
        const transmitValue = pino.levels.values[transmitLevel];
        const methodValue = pino.levels.values[level];
        if (methodValue < transmitValue) return
        transmit(this, {
          ts,
          methodLevel: level,
          methodValue,
          transmitLevel,
          transmitValue: pino.levels.values[opts.transmit.level || logger.level],
          send: opts.transmit.send,
          val: logger.levelVal
        }, args);
      }
    }
  })(logger[level]);
}

function asObject (logger, level, args, ts) {
  if (logger._serialize) applySerializers(args, logger._serialize, logger.serializers, logger._stdErrSerialize);
  const argsCloned = args.slice();
  let msg = argsCloned[0];
  const o = {};
  if (ts) {
    o.time = ts;
  }
  o.level = pino.levels.values[level];
  let lvl = (logger._childLevel | 0) + 1;
  if (lvl < 1) lvl = 1;
  // deliberate, catching objects, arrays
  if (msg !== null && typeof msg === 'object') {
    while (lvl-- && typeof argsCloned[0] === 'object') {
      Object.assign(o, argsCloned.shift());
    }
    msg = argsCloned.length ? quickFormatUnescaped(argsCloned.shift(), argsCloned) : undefined;
  } else if (typeof msg === 'string') msg = quickFormatUnescaped(argsCloned.shift(), argsCloned);
  if (msg !== undefined) o.msg = msg;
  return o
}

function applySerializers (args, serialize, serializers, stdErrSerialize) {
  for (const i in args) {
    if (stdErrSerialize && args[i] instanceof Error) {
      args[i] = pino.stdSerializers.err(args[i]);
    } else if (typeof args[i] === 'object' && !Array.isArray(args[i])) {
      for (const k in args[i]) {
        if (serialize && serialize.indexOf(k) > -1 && k in serializers) {
          args[i][k] = serializers[k](args[i][k]);
        }
      }
    }
  }
}

function bind (parent, bindings, level) {
  return function () {
    const args = new Array(1 + arguments.length);
    args[0] = bindings;
    for (var i = 1; i < args.length; i++) {
      args[i] = arguments[i - 1];
    }
    return parent[level].apply(this, args)
  }
}

function transmit (logger, opts, args) {
  const send = opts.send;
  const ts = opts.ts;
  const methodLevel = opts.methodLevel;
  const methodValue = opts.methodValue;
  const val = opts.val;
  const bindings = logger._logEvent.bindings;

  applySerializers(
    args,
    logger._serialize || Object.keys(logger.serializers),
    logger.serializers,
    logger._stdErrSerialize === undefined ? true : logger._stdErrSerialize
  );
  logger._logEvent.ts = ts;
  logger._logEvent.messages = args.filter(function (arg) {
    // bindings can only be objects, so reference equality check via indexOf is fine
    return bindings.indexOf(arg) === -1
  });

  logger._logEvent.level.label = methodLevel;
  logger._logEvent.level.value = methodValue;

  send(methodLevel, logger._logEvent, val);

  logger._logEvent = createLogEventShape(bindings);
}

function createLogEventShape (bindings) {
  return {
    ts: 0,
    messages: [],
    bindings: bindings || [],
    level: { label: '', value: 0 }
  }
}

function asErrValue (err) {
  const obj = {
    type: err.constructor.name,
    msg: err.message,
    stack: err.stack
  };
  for (const key in err) {
    if (obj[key] === undefined) {
      obj[key] = err[key];
    }
  }
  return obj
}

function getTimeFunction (opts) {
  if (typeof opts.timestamp === 'function') {
    return opts.timestamp
  }
  if (opts.timestamp === false) {
    return nullTime
  }
  return epochTime
}

function mock () { return {} }
function passthrough (a) { return a }
function noop () {}

function nullTime () { return false }
function epochTime () { return Date.now() }
function unixTime () { return Math.round(Date.now() / 1000.0) }
function isoTime () { return new Date(Date.now()).toISOString() } // using Date.now() for testability

/* eslint-disable */
/* istanbul ignore next */
function pfGlobalThisOrFallback () {
  function defd (o) { return typeof o !== 'undefined' && o }
  try {
    if (typeof globalThis !== 'undefined') return globalThis
    Object.defineProperty(Object.prototype, 'globalThis', {
      get: function () {
        delete Object.prototype.globalThis;
        return (this.globalThis = this)
      },
      configurable: true
    });
    return globalThis
  } catch (e) {
    return defd(self) || defd(window) || defd(this) || {}
  }
}

var colorName = {
	"aliceblue": [240, 248, 255],
	"antiquewhite": [250, 235, 215],
	"aqua": [0, 255, 255],
	"aquamarine": [127, 255, 212],
	"azure": [240, 255, 255],
	"beige": [245, 245, 220],
	"bisque": [255, 228, 196],
	"black": [0, 0, 0],
	"blanchedalmond": [255, 235, 205],
	"blue": [0, 0, 255],
	"blueviolet": [138, 43, 226],
	"brown": [165, 42, 42],
	"burlywood": [222, 184, 135],
	"cadetblue": [95, 158, 160],
	"chartreuse": [127, 255, 0],
	"chocolate": [210, 105, 30],
	"coral": [255, 127, 80],
	"cornflowerblue": [100, 149, 237],
	"cornsilk": [255, 248, 220],
	"crimson": [220, 20, 60],
	"cyan": [0, 255, 255],
	"darkblue": [0, 0, 139],
	"darkcyan": [0, 139, 139],
	"darkgoldenrod": [184, 134, 11],
	"darkgray": [169, 169, 169],
	"darkgreen": [0, 100, 0],
	"darkgrey": [169, 169, 169],
	"darkkhaki": [189, 183, 107],
	"darkmagenta": [139, 0, 139],
	"darkolivegreen": [85, 107, 47],
	"darkorange": [255, 140, 0],
	"darkorchid": [153, 50, 204],
	"darkred": [139, 0, 0],
	"darksalmon": [233, 150, 122],
	"darkseagreen": [143, 188, 143],
	"darkslateblue": [72, 61, 139],
	"darkslategray": [47, 79, 79],
	"darkslategrey": [47, 79, 79],
	"darkturquoise": [0, 206, 209],
	"darkviolet": [148, 0, 211],
	"deeppink": [255, 20, 147],
	"deepskyblue": [0, 191, 255],
	"dimgray": [105, 105, 105],
	"dimgrey": [105, 105, 105],
	"dodgerblue": [30, 144, 255],
	"firebrick": [178, 34, 34],
	"floralwhite": [255, 250, 240],
	"forestgreen": [34, 139, 34],
	"fuchsia": [255, 0, 255],
	"gainsboro": [220, 220, 220],
	"ghostwhite": [248, 248, 255],
	"gold": [255, 215, 0],
	"goldenrod": [218, 165, 32],
	"gray": [128, 128, 128],
	"green": [0, 128, 0],
	"greenyellow": [173, 255, 47],
	"grey": [128, 128, 128],
	"honeydew": [240, 255, 240],
	"hotpink": [255, 105, 180],
	"indianred": [205, 92, 92],
	"indigo": [75, 0, 130],
	"ivory": [255, 255, 240],
	"khaki": [240, 230, 140],
	"lavender": [230, 230, 250],
	"lavenderblush": [255, 240, 245],
	"lawngreen": [124, 252, 0],
	"lemonchiffon": [255, 250, 205],
	"lightblue": [173, 216, 230],
	"lightcoral": [240, 128, 128],
	"lightcyan": [224, 255, 255],
	"lightgoldenrodyellow": [250, 250, 210],
	"lightgray": [211, 211, 211],
	"lightgreen": [144, 238, 144],
	"lightgrey": [211, 211, 211],
	"lightpink": [255, 182, 193],
	"lightsalmon": [255, 160, 122],
	"lightseagreen": [32, 178, 170],
	"lightskyblue": [135, 206, 250],
	"lightslategray": [119, 136, 153],
	"lightslategrey": [119, 136, 153],
	"lightsteelblue": [176, 196, 222],
	"lightyellow": [255, 255, 224],
	"lime": [0, 255, 0],
	"limegreen": [50, 205, 50],
	"linen": [250, 240, 230],
	"magenta": [255, 0, 255],
	"maroon": [128, 0, 0],
	"mediumaquamarine": [102, 205, 170],
	"mediumblue": [0, 0, 205],
	"mediumorchid": [186, 85, 211],
	"mediumpurple": [147, 112, 219],
	"mediumseagreen": [60, 179, 113],
	"mediumslateblue": [123, 104, 238],
	"mediumspringgreen": [0, 250, 154],
	"mediumturquoise": [72, 209, 204],
	"mediumvioletred": [199, 21, 133],
	"midnightblue": [25, 25, 112],
	"mintcream": [245, 255, 250],
	"mistyrose": [255, 228, 225],
	"moccasin": [255, 228, 181],
	"navajowhite": [255, 222, 173],
	"navy": [0, 0, 128],
	"oldlace": [253, 245, 230],
	"olive": [128, 128, 0],
	"olivedrab": [107, 142, 35],
	"orange": [255, 165, 0],
	"orangered": [255, 69, 0],
	"orchid": [218, 112, 214],
	"palegoldenrod": [238, 232, 170],
	"palegreen": [152, 251, 152],
	"paleturquoise": [175, 238, 238],
	"palevioletred": [219, 112, 147],
	"papayawhip": [255, 239, 213],
	"peachpuff": [255, 218, 185],
	"peru": [205, 133, 63],
	"pink": [255, 192, 203],
	"plum": [221, 160, 221],
	"powderblue": [176, 224, 230],
	"purple": [128, 0, 128],
	"rebeccapurple": [102, 51, 153],
	"red": [255, 0, 0],
	"rosybrown": [188, 143, 143],
	"royalblue": [65, 105, 225],
	"saddlebrown": [139, 69, 19],
	"salmon": [250, 128, 114],
	"sandybrown": [244, 164, 96],
	"seagreen": [46, 139, 87],
	"seashell": [255, 245, 238],
	"sienna": [160, 82, 45],
	"silver": [192, 192, 192],
	"skyblue": [135, 206, 235],
	"slateblue": [106, 90, 205],
	"slategray": [112, 128, 144],
	"slategrey": [112, 128, 144],
	"snow": [255, 250, 250],
	"springgreen": [0, 255, 127],
	"steelblue": [70, 130, 180],
	"tan": [210, 180, 140],
	"teal": [0, 128, 128],
	"thistle": [216, 191, 216],
	"tomato": [255, 99, 71],
	"turquoise": [64, 224, 208],
	"violet": [238, 130, 238],
	"wheat": [245, 222, 179],
	"white": [255, 255, 255],
	"whitesmoke": [245, 245, 245],
	"yellow": [255, 255, 0],
	"yellowgreen": [154, 205, 50]
};

/* MIT license */

/* eslint-disable no-mixed-operators */


// NOTE: conversions should only return primitive values (i.e. arrays, or
//       values that give correct `typeof` results).
//       do not use box values types (i.e. Number(), String(), etc.)

const reverseKeywords = {};
for (const key of Object.keys(colorName)) {
	reverseKeywords[colorName[key]] = key;
}

const convert$1 = {
	rgb: {channels: 3, labels: 'rgb'},
	hsl: {channels: 3, labels: 'hsl'},
	hsv: {channels: 3, labels: 'hsv'},
	hwb: {channels: 3, labels: 'hwb'},
	cmyk: {channels: 4, labels: 'cmyk'},
	xyz: {channels: 3, labels: 'xyz'},
	lab: {channels: 3, labels: 'lab'},
	lch: {channels: 3, labels: 'lch'},
	hex: {channels: 1, labels: ['hex']},
	keyword: {channels: 1, labels: ['keyword']},
	ansi16: {channels: 1, labels: ['ansi16']},
	ansi256: {channels: 1, labels: ['ansi256']},
	hcg: {channels: 3, labels: ['h', 'c', 'g']},
	apple: {channels: 3, labels: ['r16', 'g16', 'b16']},
	gray: {channels: 1, labels: ['gray']}
};

var conversions = convert$1;

// Hide .channels and .labels properties
for (const model of Object.keys(convert$1)) {
	if (!('channels' in convert$1[model])) {
		throw new Error('missing channels property: ' + model);
	}

	if (!('labels' in convert$1[model])) {
		throw new Error('missing channel labels property: ' + model);
	}

	if (convert$1[model].labels.length !== convert$1[model].channels) {
		throw new Error('channel and label counts mismatch: ' + model);
	}

	const {channels, labels} = convert$1[model];
	delete convert$1[model].channels;
	delete convert$1[model].labels;
	Object.defineProperty(convert$1[model], 'channels', {value: channels});
	Object.defineProperty(convert$1[model], 'labels', {value: labels});
}

convert$1.rgb.hsl = function (rgb) {
	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;
	const min = Math.min(r, g, b);
	const max = Math.max(r, g, b);
	const delta = max - min;
	let h;
	let s;

	if (max === min) {
		h = 0;
	} else if (r === max) {
		h = (g - b) / delta;
	} else if (g === max) {
		h = 2 + (b - r) / delta;
	} else if (b === max) {
		h = 4 + (r - g) / delta;
	}

	h = Math.min(h * 60, 360);

	if (h < 0) {
		h += 360;
	}

	const l = (min + max) / 2;

	if (max === min) {
		s = 0;
	} else if (l <= 0.5) {
		s = delta / (max + min);
	} else {
		s = delta / (2 - max - min);
	}

	return [h, s * 100, l * 100];
};

convert$1.rgb.hsv = function (rgb) {
	let rdif;
	let gdif;
	let bdif;
	let h;
	let s;

	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;
	const v = Math.max(r, g, b);
	const diff = v - Math.min(r, g, b);
	const diffc = function (c) {
		return (v - c) / 6 / diff + 1 / 2;
	};

	if (diff === 0) {
		h = 0;
		s = 0;
	} else {
		s = diff / v;
		rdif = diffc(r);
		gdif = diffc(g);
		bdif = diffc(b);

		if (r === v) {
			h = bdif - gdif;
		} else if (g === v) {
			h = (1 / 3) + rdif - bdif;
		} else if (b === v) {
			h = (2 / 3) + gdif - rdif;
		}

		if (h < 0) {
			h += 1;
		} else if (h > 1) {
			h -= 1;
		}
	}

	return [
		h * 360,
		s * 100,
		v * 100
	];
};

convert$1.rgb.hwb = function (rgb) {
	const r = rgb[0];
	const g = rgb[1];
	let b = rgb[2];
	const h = convert$1.rgb.hsl(rgb)[0];
	const w = 1 / 255 * Math.min(r, Math.min(g, b));

	b = 1 - 1 / 255 * Math.max(r, Math.max(g, b));

	return [h, w * 100, b * 100];
};

convert$1.rgb.cmyk = function (rgb) {
	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;

	const k = Math.min(1 - r, 1 - g, 1 - b);
	const c = (1 - r - k) / (1 - k) || 0;
	const m = (1 - g - k) / (1 - k) || 0;
	const y = (1 - b - k) / (1 - k) || 0;

	return [c * 100, m * 100, y * 100, k * 100];
};

function comparativeDistance(x, y) {
	/*
		See https://en.m.wikipedia.org/wiki/Euclidean_distance#Squared_Euclidean_distance
	*/
	return (
		((x[0] - y[0]) ** 2) +
		((x[1] - y[1]) ** 2) +
		((x[2] - y[2]) ** 2)
	);
}

convert$1.rgb.keyword = function (rgb) {
	const reversed = reverseKeywords[rgb];
	if (reversed) {
		return reversed;
	}

	let currentClosestDistance = Infinity;
	let currentClosestKeyword;

	for (const keyword of Object.keys(colorName)) {
		const value = colorName[keyword];

		// Compute comparative distance
		const distance = comparativeDistance(rgb, value);

		// Check if its less, if so set as closest
		if (distance < currentClosestDistance) {
			currentClosestDistance = distance;
			currentClosestKeyword = keyword;
		}
	}

	return currentClosestKeyword;
};

convert$1.keyword.rgb = function (keyword) {
	return colorName[keyword];
};

convert$1.rgb.xyz = function (rgb) {
	let r = rgb[0] / 255;
	let g = rgb[1] / 255;
	let b = rgb[2] / 255;

	// Assume sRGB
	r = r > 0.04045 ? (((r + 0.055) / 1.055) ** 2.4) : (r / 12.92);
	g = g > 0.04045 ? (((g + 0.055) / 1.055) ** 2.4) : (g / 12.92);
	b = b > 0.04045 ? (((b + 0.055) / 1.055) ** 2.4) : (b / 12.92);

	const x = (r * 0.4124) + (g * 0.3576) + (b * 0.1805);
	const y = (r * 0.2126) + (g * 0.7152) + (b * 0.0722);
	const z = (r * 0.0193) + (g * 0.1192) + (b * 0.9505);

	return [x * 100, y * 100, z * 100];
};

convert$1.rgb.lab = function (rgb) {
	const xyz = convert$1.rgb.xyz(rgb);
	let x = xyz[0];
	let y = xyz[1];
	let z = xyz[2];

	x /= 95.047;
	y /= 100;
	z /= 108.883;

	x = x > 0.008856 ? (x ** (1 / 3)) : (7.787 * x) + (16 / 116);
	y = y > 0.008856 ? (y ** (1 / 3)) : (7.787 * y) + (16 / 116);
	z = z > 0.008856 ? (z ** (1 / 3)) : (7.787 * z) + (16 / 116);

	const l = (116 * y) - 16;
	const a = 500 * (x - y);
	const b = 200 * (y - z);

	return [l, a, b];
};

convert$1.hsl.rgb = function (hsl) {
	const h = hsl[0] / 360;
	const s = hsl[1] / 100;
	const l = hsl[2] / 100;
	let t2;
	let t3;
	let val;

	if (s === 0) {
		val = l * 255;
		return [val, val, val];
	}

	if (l < 0.5) {
		t2 = l * (1 + s);
	} else {
		t2 = l + s - l * s;
	}

	const t1 = 2 * l - t2;

	const rgb = [0, 0, 0];
	for (let i = 0; i < 3; i++) {
		t3 = h + 1 / 3 * -(i - 1);
		if (t3 < 0) {
			t3++;
		}

		if (t3 > 1) {
			t3--;
		}

		if (6 * t3 < 1) {
			val = t1 + (t2 - t1) * 6 * t3;
		} else if (2 * t3 < 1) {
			val = t2;
		} else if (3 * t3 < 2) {
			val = t1 + (t2 - t1) * (2 / 3 - t3) * 6;
		} else {
			val = t1;
		}

		rgb[i] = val * 255;
	}

	return rgb;
};

convert$1.hsl.hsv = function (hsl) {
	const h = hsl[0];
	let s = hsl[1] / 100;
	let l = hsl[2] / 100;
	let smin = s;
	const lmin = Math.max(l, 0.01);

	l *= 2;
	s *= (l <= 1) ? l : 2 - l;
	smin *= lmin <= 1 ? lmin : 2 - lmin;
	const v = (l + s) / 2;
	const sv = l === 0 ? (2 * smin) / (lmin + smin) : (2 * s) / (l + s);

	return [h, sv * 100, v * 100];
};

convert$1.hsv.rgb = function (hsv) {
	const h = hsv[0] / 60;
	const s = hsv[1] / 100;
	let v = hsv[2] / 100;
	const hi = Math.floor(h) % 6;

	const f = h - Math.floor(h);
	const p = 255 * v * (1 - s);
	const q = 255 * v * (1 - (s * f));
	const t = 255 * v * (1 - (s * (1 - f)));
	v *= 255;

	switch (hi) {
		case 0:
			return [v, t, p];
		case 1:
			return [q, v, p];
		case 2:
			return [p, v, t];
		case 3:
			return [p, q, v];
		case 4:
			return [t, p, v];
		case 5:
			return [v, p, q];
	}
};

convert$1.hsv.hsl = function (hsv) {
	const h = hsv[0];
	const s = hsv[1] / 100;
	const v = hsv[2] / 100;
	const vmin = Math.max(v, 0.01);
	let sl;
	let l;

	l = (2 - s) * v;
	const lmin = (2 - s) * vmin;
	sl = s * vmin;
	sl /= (lmin <= 1) ? lmin : 2 - lmin;
	sl = sl || 0;
	l /= 2;

	return [h, sl * 100, l * 100];
};

// http://dev.w3.org/csswg/css-color/#hwb-to-rgb
convert$1.hwb.rgb = function (hwb) {
	const h = hwb[0] / 360;
	let wh = hwb[1] / 100;
	let bl = hwb[2] / 100;
	const ratio = wh + bl;
	let f;

	// Wh + bl cant be > 1
	if (ratio > 1) {
		wh /= ratio;
		bl /= ratio;
	}

	const i = Math.floor(6 * h);
	const v = 1 - bl;
	f = 6 * h - i;

	if ((i & 0x01) !== 0) {
		f = 1 - f;
	}

	const n = wh + f * (v - wh); // Linear interpolation

	let r;
	let g;
	let b;
	/* eslint-disable max-statements-per-line,no-multi-spaces */
	switch (i) {
		default:
		case 6:
		case 0: r = v;  g = n;  b = wh; break;
		case 1: r = n;  g = v;  b = wh; break;
		case 2: r = wh; g = v;  b = n; break;
		case 3: r = wh; g = n;  b = v; break;
		case 4: r = n;  g = wh; b = v; break;
		case 5: r = v;  g = wh; b = n; break;
	}
	/* eslint-enable max-statements-per-line,no-multi-spaces */

	return [r * 255, g * 255, b * 255];
};

convert$1.cmyk.rgb = function (cmyk) {
	const c = cmyk[0] / 100;
	const m = cmyk[1] / 100;
	const y = cmyk[2] / 100;
	const k = cmyk[3] / 100;

	const r = 1 - Math.min(1, c * (1 - k) + k);
	const g = 1 - Math.min(1, m * (1 - k) + k);
	const b = 1 - Math.min(1, y * (1 - k) + k);

	return [r * 255, g * 255, b * 255];
};

convert$1.xyz.rgb = function (xyz) {
	const x = xyz[0] / 100;
	const y = xyz[1] / 100;
	const z = xyz[2] / 100;
	let r;
	let g;
	let b;

	r = (x * 3.2406) + (y * -1.5372) + (z * -0.4986);
	g = (x * -0.9689) + (y * 1.8758) + (z * 0.0415);
	b = (x * 0.0557) + (y * -0.2040) + (z * 1.0570);

	// Assume sRGB
	r = r > 0.0031308
		? ((1.055 * (r ** (1.0 / 2.4))) - 0.055)
		: r * 12.92;

	g = g > 0.0031308
		? ((1.055 * (g ** (1.0 / 2.4))) - 0.055)
		: g * 12.92;

	b = b > 0.0031308
		? ((1.055 * (b ** (1.0 / 2.4))) - 0.055)
		: b * 12.92;

	r = Math.min(Math.max(0, r), 1);
	g = Math.min(Math.max(0, g), 1);
	b = Math.min(Math.max(0, b), 1);

	return [r * 255, g * 255, b * 255];
};

convert$1.xyz.lab = function (xyz) {
	let x = xyz[0];
	let y = xyz[1];
	let z = xyz[2];

	x /= 95.047;
	y /= 100;
	z /= 108.883;

	x = x > 0.008856 ? (x ** (1 / 3)) : (7.787 * x) + (16 / 116);
	y = y > 0.008856 ? (y ** (1 / 3)) : (7.787 * y) + (16 / 116);
	z = z > 0.008856 ? (z ** (1 / 3)) : (7.787 * z) + (16 / 116);

	const l = (116 * y) - 16;
	const a = 500 * (x - y);
	const b = 200 * (y - z);

	return [l, a, b];
};

convert$1.lab.xyz = function (lab) {
	const l = lab[0];
	const a = lab[1];
	const b = lab[2];
	let x;
	let y;
	let z;

	y = (l + 16) / 116;
	x = a / 500 + y;
	z = y - b / 200;

	const y2 = y ** 3;
	const x2 = x ** 3;
	const z2 = z ** 3;
	y = y2 > 0.008856 ? y2 : (y - 16 / 116) / 7.787;
	x = x2 > 0.008856 ? x2 : (x - 16 / 116) / 7.787;
	z = z2 > 0.008856 ? z2 : (z - 16 / 116) / 7.787;

	x *= 95.047;
	y *= 100;
	z *= 108.883;

	return [x, y, z];
};

convert$1.lab.lch = function (lab) {
	const l = lab[0];
	const a = lab[1];
	const b = lab[2];
	let h;

	const hr = Math.atan2(b, a);
	h = hr * 360 / 2 / Math.PI;

	if (h < 0) {
		h += 360;
	}

	const c = Math.sqrt(a * a + b * b);

	return [l, c, h];
};

convert$1.lch.lab = function (lch) {
	const l = lch[0];
	const c = lch[1];
	const h = lch[2];

	const hr = h / 360 * 2 * Math.PI;
	const a = c * Math.cos(hr);
	const b = c * Math.sin(hr);

	return [l, a, b];
};

convert$1.rgb.ansi16 = function (args, saturation = null) {
	const [r, g, b] = args;
	let value = saturation === null ? convert$1.rgb.hsv(args)[2] : saturation; // Hsv -> ansi16 optimization

	value = Math.round(value / 50);

	if (value === 0) {
		return 30;
	}

	let ansi = 30
		+ ((Math.round(b / 255) << 2)
		| (Math.round(g / 255) << 1)
		| Math.round(r / 255));

	if (value === 2) {
		ansi += 60;
	}

	return ansi;
};

convert$1.hsv.ansi16 = function (args) {
	// Optimization here; we already know the value and don't need to get
	// it converted for us.
	return convert$1.rgb.ansi16(convert$1.hsv.rgb(args), args[2]);
};

convert$1.rgb.ansi256 = function (args) {
	const r = args[0];
	const g = args[1];
	const b = args[2];

	// We use the extended greyscale palette here, with the exception of
	// black and white. normal palette only has 4 greyscale shades.
	if (r === g && g === b) {
		if (r < 8) {
			return 16;
		}

		if (r > 248) {
			return 231;
		}

		return Math.round(((r - 8) / 247) * 24) + 232;
	}

	const ansi = 16
		+ (36 * Math.round(r / 255 * 5))
		+ (6 * Math.round(g / 255 * 5))
		+ Math.round(b / 255 * 5);

	return ansi;
};

convert$1.ansi16.rgb = function (args) {
	let color = args % 10;

	// Handle greyscale
	if (color === 0 || color === 7) {
		if (args > 50) {
			color += 3.5;
		}

		color = color / 10.5 * 255;

		return [color, color, color];
	}

	const mult = (~~(args > 50) + 1) * 0.5;
	const r = ((color & 1) * mult) * 255;
	const g = (((color >> 1) & 1) * mult) * 255;
	const b = (((color >> 2) & 1) * mult) * 255;

	return [r, g, b];
};

convert$1.ansi256.rgb = function (args) {
	// Handle greyscale
	if (args >= 232) {
		const c = (args - 232) * 10 + 8;
		return [c, c, c];
	}

	args -= 16;

	let rem;
	const r = Math.floor(args / 36) / 5 * 255;
	const g = Math.floor((rem = args % 36) / 6) / 5 * 255;
	const b = (rem % 6) / 5 * 255;

	return [r, g, b];
};

convert$1.rgb.hex = function (args) {
	const integer = ((Math.round(args[0]) & 0xFF) << 16)
		+ ((Math.round(args[1]) & 0xFF) << 8)
		+ (Math.round(args[2]) & 0xFF);

	const string = integer.toString(16).toUpperCase();
	return '000000'.substring(string.length) + string;
};

convert$1.hex.rgb = function (args) {
	const match = args.toString(16).match(/[a-f0-9]{6}|[a-f0-9]{3}/i);
	if (!match) {
		return [0, 0, 0];
	}

	let colorString = match[0];

	if (match[0].length === 3) {
		colorString = colorString.split('').map(char => {
			return char + char;
		}).join('');
	}

	const integer = parseInt(colorString, 16);
	const r = (integer >> 16) & 0xFF;
	const g = (integer >> 8) & 0xFF;
	const b = integer & 0xFF;

	return [r, g, b];
};

convert$1.rgb.hcg = function (rgb) {
	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;
	const max = Math.max(Math.max(r, g), b);
	const min = Math.min(Math.min(r, g), b);
	const chroma = (max - min);
	let grayscale;
	let hue;

	if (chroma < 1) {
		grayscale = min / (1 - chroma);
	} else {
		grayscale = 0;
	}

	if (chroma <= 0) {
		hue = 0;
	} else
	if (max === r) {
		hue = ((g - b) / chroma) % 6;
	} else
	if (max === g) {
		hue = 2 + (b - r) / chroma;
	} else {
		hue = 4 + (r - g) / chroma;
	}

	hue /= 6;
	hue %= 1;

	return [hue * 360, chroma * 100, grayscale * 100];
};

convert$1.hsl.hcg = function (hsl) {
	const s = hsl[1] / 100;
	const l = hsl[2] / 100;

	const c = l < 0.5 ? (2.0 * s * l) : (2.0 * s * (1.0 - l));

	let f = 0;
	if (c < 1.0) {
		f = (l - 0.5 * c) / (1.0 - c);
	}

	return [hsl[0], c * 100, f * 100];
};

convert$1.hsv.hcg = function (hsv) {
	const s = hsv[1] / 100;
	const v = hsv[2] / 100;

	const c = s * v;
	let f = 0;

	if (c < 1.0) {
		f = (v - c) / (1 - c);
	}

	return [hsv[0], c * 100, f * 100];
};

convert$1.hcg.rgb = function (hcg) {
	const h = hcg[0] / 360;
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;

	if (c === 0.0) {
		return [g * 255, g * 255, g * 255];
	}

	const pure = [0, 0, 0];
	const hi = (h % 1) * 6;
	const v = hi % 1;
	const w = 1 - v;
	let mg = 0;

	/* eslint-disable max-statements-per-line */
	switch (Math.floor(hi)) {
		case 0:
			pure[0] = 1; pure[1] = v; pure[2] = 0; break;
		case 1:
			pure[0] = w; pure[1] = 1; pure[2] = 0; break;
		case 2:
			pure[0] = 0; pure[1] = 1; pure[2] = v; break;
		case 3:
			pure[0] = 0; pure[1] = w; pure[2] = 1; break;
		case 4:
			pure[0] = v; pure[1] = 0; pure[2] = 1; break;
		default:
			pure[0] = 1; pure[1] = 0; pure[2] = w;
	}
	/* eslint-enable max-statements-per-line */

	mg = (1.0 - c) * g;

	return [
		(c * pure[0] + mg) * 255,
		(c * pure[1] + mg) * 255,
		(c * pure[2] + mg) * 255
	];
};

convert$1.hcg.hsv = function (hcg) {
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;

	const v = c + g * (1.0 - c);
	let f = 0;

	if (v > 0.0) {
		f = c / v;
	}

	return [hcg[0], f * 100, v * 100];
};

convert$1.hcg.hsl = function (hcg) {
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;

	const l = g * (1.0 - c) + 0.5 * c;
	let s = 0;

	if (l > 0.0 && l < 0.5) {
		s = c / (2 * l);
	} else
	if (l >= 0.5 && l < 1.0) {
		s = c / (2 * (1 - l));
	}

	return [hcg[0], s * 100, l * 100];
};

convert$1.hcg.hwb = function (hcg) {
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;
	const v = c + g * (1.0 - c);
	return [hcg[0], (v - c) * 100, (1 - v) * 100];
};

convert$1.hwb.hcg = function (hwb) {
	const w = hwb[1] / 100;
	const b = hwb[2] / 100;
	const v = 1 - b;
	const c = v - w;
	let g = 0;

	if (c < 1) {
		g = (v - c) / (1 - c);
	}

	return [hwb[0], c * 100, g * 100];
};

convert$1.apple.rgb = function (apple) {
	return [(apple[0] / 65535) * 255, (apple[1] / 65535) * 255, (apple[2] / 65535) * 255];
};

convert$1.rgb.apple = function (rgb) {
	return [(rgb[0] / 255) * 65535, (rgb[1] / 255) * 65535, (rgb[2] / 255) * 65535];
};

convert$1.gray.rgb = function (args) {
	return [args[0] / 100 * 255, args[0] / 100 * 255, args[0] / 100 * 255];
};

convert$1.gray.hsl = function (args) {
	return [0, 0, args[0]];
};

convert$1.gray.hsv = convert$1.gray.hsl;

convert$1.gray.hwb = function (gray) {
	return [0, 100, gray[0]];
};

convert$1.gray.cmyk = function (gray) {
	return [0, 0, 0, gray[0]];
};

convert$1.gray.lab = function (gray) {
	return [gray[0], 0, 0];
};

convert$1.gray.hex = function (gray) {
	const val = Math.round(gray[0] / 100 * 255) & 0xFF;
	const integer = (val << 16) + (val << 8) + val;

	const string = integer.toString(16).toUpperCase();
	return '000000'.substring(string.length) + string;
};

convert$1.rgb.gray = function (rgb) {
	const val = (rgb[0] + rgb[1] + rgb[2]) / 3;
	return [val / 255 * 100];
};

/*
	This function routes a model to all other models.

	all functions that are routed have a property `.conversion` attached
	to the returned synthetic function. This property is an array
	of strings, each with the steps in between the 'from' and 'to'
	color models (inclusive).

	conversions that are not possible simply are not included.
*/

function buildGraph() {
	const graph = {};
	// https://jsperf.com/object-keys-vs-for-in-with-closure/3
	const models = Object.keys(conversions);

	for (let len = models.length, i = 0; i < len; i++) {
		graph[models[i]] = {
			// http://jsperf.com/1-vs-infinity
			// micro-opt, but this is simple.
			distance: -1,
			parent: null
		};
	}

	return graph;
}

// https://en.wikipedia.org/wiki/Breadth-first_search
function deriveBFS(fromModel) {
	const graph = buildGraph();
	const queue = [fromModel]; // Unshift -> queue -> pop

	graph[fromModel].distance = 0;

	while (queue.length) {
		const current = queue.pop();
		const adjacents = Object.keys(conversions[current]);

		for (let len = adjacents.length, i = 0; i < len; i++) {
			const adjacent = adjacents[i];
			const node = graph[adjacent];

			if (node.distance === -1) {
				node.distance = graph[current].distance + 1;
				node.parent = current;
				queue.unshift(adjacent);
			}
		}
	}

	return graph;
}

function link(from, to) {
	return function (args) {
		return to(from(args));
	};
}

function wrapConversion(toModel, graph) {
	const path = [graph[toModel].parent, toModel];
	let fn = conversions[graph[toModel].parent][toModel];

	let cur = graph[toModel].parent;
	while (graph[cur].parent) {
		path.unshift(graph[cur].parent);
		fn = link(conversions[graph[cur].parent][cur], fn);
		cur = graph[cur].parent;
	}

	fn.conversion = path;
	return fn;
}

var route = function (fromModel) {
	const graph = deriveBFS(fromModel);
	const conversion = {};

	const models = Object.keys(graph);
	for (let len = models.length, i = 0; i < len; i++) {
		const toModel = models[i];
		const node = graph[toModel];

		if (node.parent === null) {
			// No possible conversion, or this node is the source model.
			continue;
		}

		conversion[toModel] = wrapConversion(toModel, graph);
	}

	return conversion;
};

const convert = {};

const models = Object.keys(conversions);

function wrapRaw(fn) {
	const wrappedFn = function (...args) {
		const arg0 = args[0];
		if (arg0 === undefined || arg0 === null) {
			return arg0;
		}

		if (arg0.length > 1) {
			args = arg0;
		}

		return fn(args);
	};

	// Preserve .conversion property if there is one
	if ('conversion' in fn) {
		wrappedFn.conversion = fn.conversion;
	}

	return wrappedFn;
}

function wrapRounded(fn) {
	const wrappedFn = function (...args) {
		const arg0 = args[0];

		if (arg0 === undefined || arg0 === null) {
			return arg0;
		}

		if (arg0.length > 1) {
			args = arg0;
		}

		const result = fn(args);

		// We're assuming the result is an array here.
		// see notice in conversions.js; don't use box types
		// in conversion functions.
		if (typeof result === 'object') {
			for (let len = result.length, i = 0; i < len; i++) {
				result[i] = Math.round(result[i]);
			}
		}

		return result;
	};

	// Preserve .conversion property if there is one
	if ('conversion' in fn) {
		wrappedFn.conversion = fn.conversion;
	}

	return wrappedFn;
}

models.forEach(fromModel => {
	convert[fromModel] = {};

	Object.defineProperty(convert[fromModel], 'channels', {value: conversions[fromModel].channels});
	Object.defineProperty(convert[fromModel], 'labels', {value: conversions[fromModel].labels});

	const routes = route(fromModel);
	const routeModels = Object.keys(routes);

	routeModels.forEach(toModel => {
		const fn = routes[toModel];

		convert[fromModel][toModel] = wrapRounded(fn);
		convert[fromModel][toModel].raw = wrapRaw(fn);
	});
});

var colorConvert = convert;

var ansiStyles = createCommonjsModule(function (module) {

const wrapAnsi16 = (fn, offset) => (...args) => {
	const code = fn(...args);
	return `\u001B[${code + offset}m`;
};

const wrapAnsi256 = (fn, offset) => (...args) => {
	const code = fn(...args);
	return `\u001B[${38 + offset};5;${code}m`;
};

const wrapAnsi16m = (fn, offset) => (...args) => {
	const rgb = fn(...args);
	return `\u001B[${38 + offset};2;${rgb[0]};${rgb[1]};${rgb[2]}m`;
};

const ansi2ansi = n => n;
const rgb2rgb = (r, g, b) => [r, g, b];

const setLazyProperty = (object, property, get) => {
	Object.defineProperty(object, property, {
		get: () => {
			const value = get();

			Object.defineProperty(object, property, {
				value,
				enumerable: true,
				configurable: true
			});

			return value;
		},
		enumerable: true,
		configurable: true
	});
};

/** @type {typeof import('color-convert')} */
let colorConvert$1;
const makeDynamicStyles = (wrap, targetSpace, identity, isBackground) => {
	if (colorConvert$1 === undefined) {
		colorConvert$1 = colorConvert;
	}

	const offset = isBackground ? 10 : 0;
	const styles = {};

	for (const [sourceSpace, suite] of Object.entries(colorConvert$1)) {
		const name = sourceSpace === 'ansi16' ? 'ansi' : sourceSpace;
		if (sourceSpace === targetSpace) {
			styles[name] = wrap(identity, offset);
		} else if (typeof suite === 'object') {
			styles[name] = wrap(suite[targetSpace], offset);
		}
	}

	return styles;
};

function assembleStyles() {
	const codes = new Map();
	const styles = {
		modifier: {
			reset: [0, 0],
			// 21 isn't widely supported and 22 does the same thing
			bold: [1, 22],
			dim: [2, 22],
			italic: [3, 23],
			underline: [4, 24],
			inverse: [7, 27],
			hidden: [8, 28],
			strikethrough: [9, 29]
		},
		color: {
			black: [30, 39],
			red: [31, 39],
			green: [32, 39],
			yellow: [33, 39],
			blue: [34, 39],
			magenta: [35, 39],
			cyan: [36, 39],
			white: [37, 39],

			// Bright color
			blackBright: [90, 39],
			redBright: [91, 39],
			greenBright: [92, 39],
			yellowBright: [93, 39],
			blueBright: [94, 39],
			magentaBright: [95, 39],
			cyanBright: [96, 39],
			whiteBright: [97, 39]
		},
		bgColor: {
			bgBlack: [40, 49],
			bgRed: [41, 49],
			bgGreen: [42, 49],
			bgYellow: [43, 49],
			bgBlue: [44, 49],
			bgMagenta: [45, 49],
			bgCyan: [46, 49],
			bgWhite: [47, 49],

			// Bright color
			bgBlackBright: [100, 49],
			bgRedBright: [101, 49],
			bgGreenBright: [102, 49],
			bgYellowBright: [103, 49],
			bgBlueBright: [104, 49],
			bgMagentaBright: [105, 49],
			bgCyanBright: [106, 49],
			bgWhiteBright: [107, 49]
		}
	};

	// Alias bright black as gray (and grey)
	styles.color.gray = styles.color.blackBright;
	styles.bgColor.bgGray = styles.bgColor.bgBlackBright;
	styles.color.grey = styles.color.blackBright;
	styles.bgColor.bgGrey = styles.bgColor.bgBlackBright;

	for (const [groupName, group] of Object.entries(styles)) {
		for (const [styleName, style] of Object.entries(group)) {
			styles[styleName] = {
				open: `\u001B[${style[0]}m`,
				close: `\u001B[${style[1]}m`
			};

			group[styleName] = styles[styleName];

			codes.set(style[0], style[1]);
		}

		Object.defineProperty(styles, groupName, {
			value: group,
			enumerable: false
		});
	}

	Object.defineProperty(styles, 'codes', {
		value: codes,
		enumerable: false
	});

	styles.color.close = '\u001B[39m';
	styles.bgColor.close = '\u001B[49m';

	setLazyProperty(styles.color, 'ansi', () => makeDynamicStyles(wrapAnsi16, 'ansi16', ansi2ansi, false));
	setLazyProperty(styles.color, 'ansi256', () => makeDynamicStyles(wrapAnsi256, 'ansi256', ansi2ansi, false));
	setLazyProperty(styles.color, 'ansi16m', () => makeDynamicStyles(wrapAnsi16m, 'rgb', rgb2rgb, false));
	setLazyProperty(styles.bgColor, 'ansi', () => makeDynamicStyles(wrapAnsi16, 'ansi16', ansi2ansi, true));
	setLazyProperty(styles.bgColor, 'ansi256', () => makeDynamicStyles(wrapAnsi256, 'ansi256', ansi2ansi, true));
	setLazyProperty(styles.bgColor, 'ansi16m', () => makeDynamicStyles(wrapAnsi16m, 'rgb', rgb2rgb, true));

	return styles;
}

// Make the export immutable
Object.defineProperty(module, 'exports', {
	enumerable: true,
	get: assembleStyles
});
});

var browser = {
	stdout: false,
	stderr: false
};

const stringReplaceAll$1 = (string, substring, replacer) => {
	let index = string.indexOf(substring);
	if (index === -1) {
		return string;
	}

	const substringLength = substring.length;
	let endIndex = 0;
	let returnValue = '';
	do {
		returnValue += string.substr(endIndex, index - endIndex) + substring + replacer;
		endIndex = index + substringLength;
		index = string.indexOf(substring, endIndex);
	} while (index !== -1);

	returnValue += string.substr(endIndex);
	return returnValue;
};

const stringEncaseCRLFWithFirstIndex$1 = (string, prefix, postfix, index) => {
	let endIndex = 0;
	let returnValue = '';
	do {
		const gotCR = string[index - 1] === '\r';
		returnValue += string.substr(endIndex, (gotCR ? index - 1 : index) - endIndex) + prefix + (gotCR ? '\r\n' : '\n') + postfix;
		endIndex = index + 1;
		index = string.indexOf('\n', endIndex);
	} while (index !== -1);

	returnValue += string.substr(endIndex);
	return returnValue;
};

var util = {
	stringReplaceAll: stringReplaceAll$1,
	stringEncaseCRLFWithFirstIndex: stringEncaseCRLFWithFirstIndex$1
};

const TEMPLATE_REGEX = /(?:\\(u(?:[a-f\d]{4}|\{[a-f\d]{1,6}\})|x[a-f\d]{2}|.))|(?:\{(~)?(\w+(?:\([^)]*\))?(?:\.\w+(?:\([^)]*\))?)*)(?:[ \t]|(?=\r?\n)))|(\})|((?:.|[\r\n\f])+?)/gi;
const STYLE_REGEX = /(?:^|\.)(\w+)(?:\(([^)]*)\))?/g;
const STRING_REGEX = /^(['"])((?:\\.|(?!\1)[^\\])*)\1$/;
const ESCAPE_REGEX = /\\(u(?:[a-f\d]{4}|{[a-f\d]{1,6}})|x[a-f\d]{2}|.)|([^\\])/gi;

const ESCAPES = new Map([
	['n', '\n'],
	['r', '\r'],
	['t', '\t'],
	['b', '\b'],
	['f', '\f'],
	['v', '\v'],
	['0', '\0'],
	['\\', '\\'],
	['e', '\u001B'],
	['a', '\u0007']
]);

function unescape(c) {
	const u = c[0] === 'u';
	const bracket = c[1] === '{';

	if ((u && !bracket && c.length === 5) || (c[0] === 'x' && c.length === 3)) {
		return String.fromCharCode(parseInt(c.slice(1), 16));
	}

	if (u && bracket) {
		return String.fromCodePoint(parseInt(c.slice(2, -1), 16));
	}

	return ESCAPES.get(c) || c;
}

function parseArguments(name, arguments_) {
	const results = [];
	const chunks = arguments_.trim().split(/\s*,\s*/g);
	let matches;

	for (const chunk of chunks) {
		const number = Number(chunk);
		if (!Number.isNaN(number)) {
			results.push(number);
		} else if ((matches = chunk.match(STRING_REGEX))) {
			results.push(matches[2].replace(ESCAPE_REGEX, (m, escape, character) => escape ? unescape(escape) : character));
		} else {
			throw new Error(`Invalid Chalk template style argument: ${chunk} (in style '${name}')`);
		}
	}

	return results;
}

function parseStyle(style) {
	STYLE_REGEX.lastIndex = 0;

	const results = [];
	let matches;

	while ((matches = STYLE_REGEX.exec(style)) !== null) {
		const name = matches[1];

		if (matches[2]) {
			const args = parseArguments(name, matches[2]);
			results.push([name].concat(args));
		} else {
			results.push([name]);
		}
	}

	return results;
}

function buildStyle(chalk, styles) {
	const enabled = {};

	for (const layer of styles) {
		for (const style of layer.styles) {
			enabled[style[0]] = layer.inverse ? null : style.slice(1);
		}
	}

	let current = chalk;
	for (const [styleName, styles] of Object.entries(enabled)) {
		if (!Array.isArray(styles)) {
			continue;
		}

		if (!(styleName in current)) {
			throw new Error(`Unknown Chalk style: ${styleName}`);
		}

		current = styles.length > 0 ? current[styleName](...styles) : current[styleName];
	}

	return current;
}

var templates = (chalk, temporary) => {
	const styles = [];
	const chunks = [];
	let chunk = [];

	// eslint-disable-next-line max-params
	temporary.replace(TEMPLATE_REGEX, (m, escapeCharacter, inverse, style, close, character) => {
		if (escapeCharacter) {
			chunk.push(unescape(escapeCharacter));
		} else if (style) {
			const string = chunk.join('');
			chunk = [];
			chunks.push(styles.length === 0 ? string : buildStyle(chalk, styles)(string));
			styles.push({inverse, styles: parseStyle(style)});
		} else if (close) {
			if (styles.length === 0) {
				throw new Error('Found extraneous } in Chalk template literal');
			}

			chunks.push(buildStyle(chalk, styles)(chunk.join('')));
			chunk = [];
			styles.pop();
		} else {
			chunk.push(character);
		}
	});

	chunks.push(chunk.join(''));

	if (styles.length > 0) {
		const errMessage = `Chalk template literal is missing ${styles.length} closing bracket${styles.length === 1 ? '' : 's'} (\`}\`)`;
		throw new Error(errMessage);
	}

	return chunks.join('');
};

const {stdout: stdoutColor, stderr: stderrColor} = browser;
const {
	stringReplaceAll,
	stringEncaseCRLFWithFirstIndex
} = util;

const {isArray} = Array;

// `supportsColor.level` → `ansiStyles.color[name]` mapping
const levelMapping = [
	'ansi',
	'ansi',
	'ansi256',
	'ansi16m'
];

const styles = Object.create(null);

const applyOptions = (object, options = {}) => {
	if (options.level && !(Number.isInteger(options.level) && options.level >= 0 && options.level <= 3)) {
		throw new Error('The `level` option should be an integer from 0 to 3');
	}

	// Detect level if not set manually
	const colorLevel = stdoutColor ? stdoutColor.level : 0;
	object.level = options.level === undefined ? colorLevel : options.level;
};

class ChalkClass {
	constructor(options) {
		// eslint-disable-next-line no-constructor-return
		return chalkFactory(options);
	}
}

const chalkFactory = options => {
	const chalk = {};
	applyOptions(chalk, options);

	chalk.template = (...arguments_) => chalkTag(chalk.template, ...arguments_);

	Object.setPrototypeOf(chalk, Chalk.prototype);
	Object.setPrototypeOf(chalk.template, chalk);

	chalk.template.constructor = () => {
		throw new Error('`chalk.constructor()` is deprecated. Use `new chalk.Instance()` instead.');
	};

	chalk.template.Instance = ChalkClass;

	return chalk.template;
};

function Chalk(options) {
	return chalkFactory(options);
}

for (const [styleName, style] of Object.entries(ansiStyles)) {
	styles[styleName] = {
		get() {
			const builder = createBuilder(this, createStyler(style.open, style.close, this._styler), this._isEmpty);
			Object.defineProperty(this, styleName, {value: builder});
			return builder;
		}
	};
}

styles.visible = {
	get() {
		const builder = createBuilder(this, this._styler, true);
		Object.defineProperty(this, 'visible', {value: builder});
		return builder;
	}
};

const usedModels = ['rgb', 'hex', 'keyword', 'hsl', 'hsv', 'hwb', 'ansi', 'ansi256'];

for (const model of usedModels) {
	styles[model] = {
		get() {
			const {level} = this;
			return function (...arguments_) {
				const styler = createStyler(ansiStyles.color[levelMapping[level]][model](...arguments_), ansiStyles.color.close, this._styler);
				return createBuilder(this, styler, this._isEmpty);
			};
		}
	};
}

for (const model of usedModels) {
	const bgModel = 'bg' + model[0].toUpperCase() + model.slice(1);
	styles[bgModel] = {
		get() {
			const {level} = this;
			return function (...arguments_) {
				const styler = createStyler(ansiStyles.bgColor[levelMapping[level]][model](...arguments_), ansiStyles.bgColor.close, this._styler);
				return createBuilder(this, styler, this._isEmpty);
			};
		}
	};
}

const proto = Object.defineProperties(() => {}, {
	...styles,
	level: {
		enumerable: true,
		get() {
			return this._generator.level;
		},
		set(level) {
			this._generator.level = level;
		}
	}
});

const createStyler = (open, close, parent) => {
	let openAll;
	let closeAll;
	if (parent === undefined) {
		openAll = open;
		closeAll = close;
	} else {
		openAll = parent.openAll + open;
		closeAll = close + parent.closeAll;
	}

	return {
		open,
		close,
		openAll,
		closeAll,
		parent
	};
};

const createBuilder = (self, _styler, _isEmpty) => {
	const builder = (...arguments_) => {
		if (isArray(arguments_[0]) && isArray(arguments_[0].raw)) {
			// Called as a template literal, for example: chalk.red`2 + 3 = {bold ${2+3}}`
			return applyStyle(builder, chalkTag(builder, ...arguments_));
		}

		// Single argument is hot path, implicit coercion is faster than anything
		// eslint-disable-next-line no-implicit-coercion
		return applyStyle(builder, (arguments_.length === 1) ? ('' + arguments_[0]) : arguments_.join(' '));
	};

	// We alter the prototype because we must return a function, but there is
	// no way to create a function with a different prototype
	Object.setPrototypeOf(builder, proto);

	builder._generator = self;
	builder._styler = _styler;
	builder._isEmpty = _isEmpty;

	return builder;
};

const applyStyle = (self, string) => {
	if (self.level <= 0 || !string) {
		return self._isEmpty ? '' : string;
	}

	let styler = self._styler;

	if (styler === undefined) {
		return string;
	}

	const {openAll, closeAll} = styler;
	if (string.indexOf('\u001B') !== -1) {
		while (styler !== undefined) {
			// Replace any instances already present with a re-opening code
			// otherwise only the part of the string until said closing code
			// will be colored, and the rest will simply be 'plain'.
			string = stringReplaceAll(string, styler.close, styler.open);

			styler = styler.parent;
		}
	}

	// We can move both next actions out of loop, because remaining actions in loop won't have
	// any/visible effect on parts we add here. Close the styling before a linebreak and reopen
	// after next line to fix a bleed issue on macOS: https://github.com/chalk/chalk/pull/92
	const lfIndex = string.indexOf('\n');
	if (lfIndex !== -1) {
		string = stringEncaseCRLFWithFirstIndex(string, closeAll, openAll, lfIndex);
	}

	return openAll + string + closeAll;
};

let template;
const chalkTag = (chalk, ...strings) => {
	const [firstString] = strings;

	if (!isArray(firstString) || !isArray(firstString.raw)) {
		// If chalk() was called by itself or with a string,
		// return the string itself as a string.
		return strings.join(' ');
	}

	const arguments_ = strings.slice(1);
	const parts = [firstString.raw[0]];

	for (let i = 1; i < firstString.length; i++) {
		parts.push(
			String(arguments_[i - 1]).replace(/[{}\\]/g, '\\$&'),
			String(firstString.raw[i])
		);
	}

	if (template === undefined) {
		template = templates;
	}

	return template(chalk, parts.join(''));
};

Object.defineProperties(Chalk.prototype, styles);

const chalk = Chalk(); // eslint-disable-line new-cap
chalk.supportsColor = stdoutColor;
chalk.stderr = Chalk({level: stderrColor ? stderrColor.level : 0}); // eslint-disable-line new-cap
chalk.stderr.supportsColor = stderrColor;

var source = chalk;

function getLogLevel () {
  {
    return window.localStorage.getItem('LOG') || 'info'
  }
}

var log = function createLogger (opts) {
  const defaultOpts = {
    level: getLogLevel(),
    prettifier: createPrettifier,
    browser: {
      write: browserWrite
    }
  };

  defaultOpts.prettyPrint = {};

  const logger = browser$1({
    ...defaultOpts,
    ...opts
  });
  return logger
};

function browserWrite (obj) {
  let method = convertLogNumber(obj).name;
  if (method === 'fatal') method = 'error';
  if (method === 'trace') method = 'debug';
  if (!console[method]) method = 'log';
  const prettified = prettify(obj);
  console[method](prettified);
}

function createPrettifier () {
  return prettify
}

function prettify (obj) {
  if (typeof obj === 'string') {
    obj = { message: obj };
  }
  if (!obj.message && obj.msg) obj.message = obj.msg;
  if (obj.err) obj.error = obj.err;
  const sp = ' ';

  const date = formatDate();
  const level = formatLevel(obj);

  if (obj.error && typeof obj.error === 'object' && obj.error.message) {
    if (obj.message) obj.message += source.gray(' (' + obj.error.message + ')');
    else obj.message = obj.error.message;
  }

  let out = '[';
  if (date) out += date + sp;
  out += level;
  if (obj.name) out += sp + source.gray(obj.name);
  if (obj.namespace) out += sp + source.gray(obj.namespace);
  if (obj.collection) out += sp + source.blue(collectionLabel(obj.collection));
  out += '] ';
  if (obj.res && obj.req) out += formatHttp(obj) + sp;
  out += obj.message;
  if (obj.record) out += sp + formatRecord(obj.record);
  return out + '\n'
}

function formatDate () {
  const date = new Date();
  const z = date.getTimezoneOffset() * 60 * 1000;
  const dateLocal = new Date(date - z);
  return dateLocal.toISOString().substring(0, 19)
}

function formatRecord (record) {
  let out = '';
  out += source.blue.bold(record.id);
  out += ' ' + source.blue(record.type);
  const seq = record.seq || '?';
  out += ' ' + source.gray(`${prettyHash(record.key)}@${seq}`);
  return source.blue('Record(') + out + source.blue(')')
}

function collectionLabel (collection) {
  return collection._keyOrName || collection.name || (collection.key && prettyHash(collection.key)) || '<unknown>'
}

function formatLevel (obj) {
  const { name, color } = convertLogNumber(obj);
  let format = source[color];
  if (obj.level >= 40) format = source[color].bold;
  return format(name.toUpperCase())
}

function formatHttp (obj) {
  if (!obj.res) return
  const { res, req } = obj;
  let color;
  if (res.statusCode > 200) color = 'red';
  else color = 'green';
  let out = '';
  out += source.bold(req.method);
  out += ' ' + req.url;
  out += ' ' + source.gray(obj.responseTime + 'ms');
  out += ' ' + source[color].bold(res.statusCode);
  const size = res.headers['content-length'];
  if (size) out += ' ' + source.gray(prettyBytes(Number(size)).replace(' ', ''));
  return out
}

function convertLogNumber (obj) {
  if (obj.level === 10) return { name: 'trace', color: 'gray' }
  if (obj.level === 20) return { name: 'debug', color: 'gray' }
  if (obj.level === 30) return { name: 'info', color: 'green' }
  if (obj.level === 40) return { name: 'warn', color: 'yellow' }
  if (obj.level === 50) return { name: 'error', color: 'red' }
  if (obj.level === 60) return { name: 'fatal', color: 'red' }
}

const { Record, RecordVersion } = records;



var common = { Schema: schema, Store: store, Record, RecordVersion, Type: type, Logger: log };

var DEFAULT_URL = 'http://localhost:9191/api/v1/default';
var DEFAULT_COLLECTION$1 = 'default';
var SCHEMA_RESOURCE$1 = 'sonar/resource';
var METADATA_ID$1 = 'sonar.id';
var HYPERDRIVE_SCHEME$2 = 'hyper:';

// @deprecated
var DEFAULT_ENDPOINT$1 = 'http://localhost:9191/api/v1/default';
// @deprecated
var DEFAULT_WORKSPACE$1 = 'default';

var constants = {
	DEFAULT_URL: DEFAULT_URL,
	DEFAULT_COLLECTION: DEFAULT_COLLECTION$1,
	SCHEMA_RESOURCE: SCHEMA_RESOURCE$1,
	METADATA_ID: METADATA_ID$1,
	HYPERDRIVE_SCHEME: HYPERDRIVE_SCHEME$2,
	DEFAULT_ENDPOINT: DEFAULT_ENDPOINT$1,
	DEFAULT_WORKSPACE: DEFAULT_WORKSPACE$1
};

const isNode = typeof window === 'undefined';
const parse = isNode ? require$$0$2.parse : browserParse;

const SCHEME_REGEX = /[a-z]+:\/\//i;
//                   1          2      3        4
const VERSION_REGEX = /^(dat:\/\/)?([^/]+)(\+[^/]+)(.*)$/i;

var parseDatUrl = function parseDatURL (str, parseQS) {
  // prepend the scheme if it's missing
  if (!SCHEME_REGEX.test(str)) {
    str = 'dat://' + str;
  }

  var parsed, version = null, match = VERSION_REGEX.exec(str);
  if (match) {
    // run typical parse with version segment removed
    parsed = parse((match[1] || '') + (match[2] || '') + (match[4] || ''), parseQS);
    version = match[3].slice(1);
  } else {
    parsed = parse(str, parseQS);
  }
  if (isNode) parsed.href = str; // overwrite href to include actual original
  else parsed.path = parsed.pathname; // to match node
  if (!parsed.query && parsed.searchParams) {
    parsed.query = Object.fromEntries(parsed.searchParams); // to match node
  }
  parsed.version = version; // add version segment
  return parsed
};

function browserParse (str) {
  return new URL(str)
}

const { HYPERDRIVE_SCHEME: HYPERDRIVE_SCHEME$1 } = constants;


/**
 * File system for a collection.
 */
class Fs {
  /**
   * File system for a collection.
   *
   * @constructor
   * @param {Collection} collection - Collection
   */
  constructor (collection) {
    this.endpoint = collection.endpoint + '/fs';
    this.collection = collection;
  }

  async fetch (path, opts = {}) {
    path = this.resolveURL(path);
    return this.collection.fetch(path, opts)
  }

  /**
   * Resolve a file URL into a HTTP url.
   *
   * @param {string} path - Either a hyper:// URL or a path in the collection's file system.
   * @throws Throws if the URL cannot be resolved.
   * @return {string} The HTTP URL to the file.
   */
  resolveURL (path) {
    // Support hyper:// URLs.
    if (path.startsWith(HYPERDRIVE_SCHEME$1)) {
      const url = parseDatUrl(path);
      path = url.host + url.path;
    // Support no other absolute URLs for now.
    } else if (path.indexOf('://') !== -1) {
      throw new Error('Invalid path: Only hyper:// URLs or paths are supported')
    }
    // Assume it's a path to a file in the island fs, will fail if not found.
    if (!path.startsWith('/')) path = '/' + path;
    let link = this.endpoint + path;
    if (this.collection.workspace.token) {
      link += `?token=${this.collection.workspace.token}`;
    }
    return link
  }

  /**
   * List the drives that are part of this collection.
   *
   * @async
   * @return {Promise<Array<object>>} Array of drive objects with keys `{ alias, key, writable }`
   */
  async listDrives () {
    // TODO: Move route under /fs somehow. Maybe HEAD on /?
    return this.collection.fetch('/fs-info')
  }

  /**
   * Get the contents of a directory.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {object} [opts] - Options
   * @param {string} [opts.includeStats=true] - Include metadata for each file
   * @throws Will throw if the path is not found or is not a directory.
   * @return {Promise<Array<object>>} An array of objects with file metadata.
   */
  async readdir (path, opts = {}) {
    const self = this;
    path = path || '/';
    if (path === '/') {
      const drives = await this.listDrives();
      return drives.map(drive => ({
        name: drive.alias || drive.key,
        link: '',
        resource: null,
        length: null,
        directory: true
      }))
    }
    if (path.length > 2 && path.charAt(0) === '/') path = path.substring(1);
    const alias = path.split('/')[0];

    let files = await this.fetch(path);

    if (files && files.length) {
      files = files.map(file => {
        file.link = makeLink(file);
        file.resource = getResource(file);
        return file
      });
    }
    return files

    function makeLink (file) {
      let link = `${self.endpoint}/${alias}/${file.path}`;
      if (self.collection.workspace.token) {
        link += `?token=${self.collection.workspace.token}`;
      }
      return link
    }

    function getResource (file) {
      if (!file || !file.metadata || !file.metadata['sonar.id']) return null
      return file.metadata['sonar.id']
    }
  }

  /**
   * Write a file
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {Stream|Buffer} file - File content to write. Can be either a buffer or a read stream.
   * @param {object} [opts] - Options.
   * @throws Will throw if the path cannot be written to.
   * @return {Promise}
   */
  async writeFile (path, file, opts = {}) {
    const requestType = opts.requestType || 'buffer';
    const params = {};
    if (opts.metadata) params.metadata = JSON.stringify(opts.metadata);

    return this.fetch(path, {
      method: 'PUT',
      body: file,
      params,
      responseType: 'text',
      requestType,
      onUploadProgress: opts.onUploadProgress
    })
  }

  /**
   * Read a file into a buffer.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {object} [opts] - Options. TODO: document.
   * @throws Will throw if the path is not found.
   * @return {Promise<ArrayBuffer|Buffer>} The file content. A Buffer object in Node.js, a ArrayBuffer object in the browser.
   */
  async readFile (path, opts = {}) {
    opts.responseType = opts.responseType || 'buffer';
    opts.requestType = opts.requestType || 'buffer';
    return this.fetch(path, opts)
  }

  /**
   * Get a read stream for a file.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {object} [opts] - Options. TODO: document.
   * @throws Will throw if the path is not found.
   * @return {Promise<ReadableStream|Readable>} A `stream.Readable` in Node.js, a `ReadableStream`in the browser.
   */
  async createReadStream (path, opts = {}) {
    opts.responseType = 'stream';
    return this.readFile(path, opts)
  }

  /**
   * Get metadata about a file.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @throws Will throw if the path is not found.
   * @return {Promise<object>} A plain object with the stat info.
   */
  async statFile (path) {
    return this.fetch(path)
  }
}

var fs = Fs;

const { SCHEMA_RESOURCE, HYPERDRIVE_SCHEME, METADATA_ID } = constants;

var resources = class Resources {
  constructor (collection) {
    this.collection = collection;
  }

  async _localDriveKey () {
    if (this._localDrive) return this._localDrive
    const drives = await this.collection.fs.listDrives();
    const writableDrives = drives.filter(f => f.writable);
    if (!writableDrives.length) throw new Error('No writable drive')
    this._localDrive = writableDrives[0].key;
    return this._localDrive
  }

  async writeFile (record, file, opts = {}) {
    const url = getContentUrl(record);
    if (!url) throw new Error('record has no file url')
    // if (!opts.metadata) opts.metadata = {}
    // opts.metadata[METADATA_ID] = record.id
    return this.collection.fs.writeFile(url, file, opts)
  }

  async readFile (record, opts = {}) {
    const url = getContentUrl(record);
    if (!url) throw new Error('record has no file url')
    return this.collection.fs.readFile(url, opts)
  }

  // TODO: Remove
  async create (value, opts) {
    return this.prepare(value, opts)
  }

  // TODO: Remove
  async prepare (value, opts = {}) {
    const { filename, prefix } = value;
    if (!filename) throw new Error('Filename is required')
    if (filename.indexOf('/') !== -1) throw new Error('Invalid filename')
    if (opts.scoped) throw new Error('Scoped option is not supported')

    let filepath;
    if (prefix) filepath = [prefix, filename].join('/');
    else filepath = filename;

    const drivekey = await this._localDriveKey();
    const contentUrl = createHyperdriveUrl(drivekey, filepath);

    let id;
    // TODO: Check for resources also/instead?
    // This checks only the local drive.
    try {
      var existing = await this.collection.fs.statFile(contentUrl);
    } catch (err) {}

    if (existing) {
      id = existing.metadata[METADATA_ID];
      if (!id) {
        if (!opts.force) throw new Error('file exists and has no resource attached. set fore to overwrite.')
      } else {
        // TODO: Preserve fields from an old resource?
        // const oldResource = await this.get({ id: existing.metadata[METADATA_ID] })
        if (!opts.update) throw new Error(`file exists, with resource ${id}. set update to overwrite.`)
      }
    }

    id = id || opts.id || Math.random();

    const record = {
      type: SCHEMA_RESOURCE,
      id,
      value: {
        ...value,
        contentUrl,
        filename
      }
    };

    return this.collection.schema.Record(record)

    // const res = await this.collection.put(record)
    // // TODO: This should get by keyseq. Or put should just return the
    // // putted record.
    // const records = await this.collection.get({
    //   id: res.id,
    //   type: SCHEMA_RESOURCE
    // }, { waitForSync: true })
    // if (!records.length) {
    //   throw new Error('error loading created resource')
    // }
    // return records[0]
  }

  resolveFileURL (record) {
    const url = getContentUrl(record);
    return this.collection.fs.resolveURL(url)
  }
};

function createHyperdriveUrl (key, path) {
  return `${HYPERDRIVE_SCHEME}//${key}/${path}`
}

function getContentUrl (record) {
  if (!record.hasType(SCHEMA_RESOURCE)) return null
  if (!record.value.contentUrl) return null
  return record.value.contentUrl
}

const debug$1 = require$$0('sonar-client');
const { Readable, Writable, Transform } = streamx;
const { EventEmitter: EventEmitter$1 } = require$$0$1;

const { Schema, Store } = common;



function uuid () {
  return base32Encode(randomBytes(16), 'Crockford').toLowerCase()
}

class Collection extends EventEmitter$1 {
  static uuid () {
    return uuid()
  }

  /**
   * Remote collection
   *
   * @constructor
   * @param {Workspace} workspace - Remote workspace
   * @param {string} nameOrKey - Name or key of the collection
   * @return {Collection}
   */
  constructor (workspace, nameOrKey) {
    super();
    this.endpoint = workspace.endpoint + '/collection/' + nameOrKey;
    this.workspace = workspace;
    this._info = {};
    this._nameOrKey = nameOrKey;
    this._eventStreams = new Set();

    this.fs = new fs(this);
    this.resources = new resources(this);
    this.log = workspace.log.child({ collection: this });
    this.setMaxListeners(256);
  }

  get name () {
    if (this._info) return this._info.name
    return this._nameOrKey
  }

  get key () {
    return this._info && this._info.key
  }

  get localKey () {
    return this._info && this._info.localKey
  }

  get info () {
    return this._info
  }

  get id () {
    return this._info && this._info.id
  }

  get length () {
    return this._length || this._info.length || 0
  }

  /**
   * Populate info and schemas for this collection from server.
   *
   * @async
   * @throws Will throw if this collection does not exist or cannot be accessed.
   * @return {Promise<void>}
   */
  async open (reset = false) {
    if (!this._openPromise || reset) this._openPromise = this._open();
    return this._openPromise
  }

  async _open () {
    this._info = await this.fetch('/');

    this.schema = new Schema({ defaultNamespace: this.id });
    const typeSpecs = await this.fetch('/schema');
    for (const typeSpec of Object.values(typeSpecs)) {
      this.schema.addType(typeSpec);
    }
    this.store = new Store({ schema: this.schema });
    this.opened = true;
    this.emit('open');
  }

  /**
   * Close the collection.
   *
   * Properly closes open HTTP requests.
   * @async
   */
  async close () {
    for (const stream of this._eventStreams) {
      stream.destroy();
    }
    if (this._eventSource) this._eventSource.close();
  }

  /**
   * Put a new feed to the collection.
   *
   * @async
   * @param {string} key - The hex-encoded key of the feed to add.
   * @param {object} [info] - Optional information about the feed.
   *                          TODO: Document
   */
  async putFeed (key, info = {}) {
    return this.fetch('/feed/' + key, {
      method: 'PUT',
      body: info
    })
  }

  async addFeed (key, info = {}) {
    return this.putFeed(key, info)
  }

  /**
   * Query the collection.
   *
   * Returns an array of matching records.
   *
   * Records may have a meta property that includes query-specific metadata (e.g. the score for search queries).
   *
   * @async
   * @param {string} name - The name of a supported query.
   *                        Supported queries that ship with @arsonar/core are:
   *                        records, search, relations, history and indexes.
   * @param {object} args - The arguments for the query. Depends on the query being used.
   *                        For records: `{ schema, name, id }`
   *                        For history: `{ from: timestamp, to: timestamp }`
   *                        For search: Either a "string" for a simple full-text search, or a
   *                           tantivy query object.
   *                        For indexes: `{ schema, prop, value, from, to, reverse, limit }`
   *                           (to be documented)
   *                        For relations: `{ subject, object, predicate }`
   *                           where subject and object are ids and predicate is `type#field`
   *
   * @param {object} [opts] - Optional options
   * @param {boolean} [opts.sync=false] Wait for all pending indexing operations to be finished.
   *
   * @return {Promise<Array<Record>>} A promise that resolves to an array of record objects.
   */
  async query (name, args, opts) {
    if (this._cacheid) {
      opts.cacheid = this._cacheid;
    }

    let records = await this.fetch('/query/' + name, {
      method: 'POST',
      body: args,
      params: opts
    });

    for (const key of records.keys()) {
      const record = records[key];
      try {
        records[key] = this.store.cacheRecord(record);
      } catch (err) {
        // TODO: Where should these errors go
        console.error('Error when upcasting record', err);
        console.error('Record: ', record);
        records[key] = null;
      }
    }
    records = records.filter(r => r);

    // if (this._cacheid) {
    //   return this._cache.batch(records)
    // }

    return records
  }

  /**
   * Put a new record into the collection.
   *
   * @async
   * @param {object} record - The record.
   * @param {string} record.schema - The schema of the record.
   * @param {string} [record.id] - The entity id of the record. If empoty an id will be created.
   * @param {object} record.value - Value of the record.
   * @throws Throws if the record is invalid.
   * @return {Promise<object>} An object with an `{ id }` property.
   */
  async put (record) {
    // If the record has no id set (= is a new record), generate and add a random id.
    if (!record.id) record.id = uuid();
    // This checks if the record has type, id, value set and if the type is present
    // in the collection's schema. Throws an error if not.
    // TODO: Add feature to @arsonar/common to validate the record's value against the schema.
    record = this.schema.RecordVersion(record);
    const resultRecord = await this.fetch('/', {
      method: 'POST',
      body: record
    });
    return this.store.cacheRecord(resultRecord)
  }

  /**
   * Get records by their semantic address (type and id) or by their storage address (key and seq).
   *
   * @async
   * @param {object} req - The get request. Either `{ type, id }` or `{ key, seq }`.
   * @param {object} [opts] - Optional options.
   * @param {boolean} [opts.sync=false] Wait for all pending indexing operations to be finished.
   * @return {Promise<Array<object>>} A promise that resolves to an array of record objects.
   */
  async get (req, opts) {
    // TODO: Implement RecordCache.has
    // if (this._cache.has(req)) {
    //   return this._cache.get(req)
    // }
    return this.query('records', req, opts)
  }

  /**
   * Get a specific version of a record.
   *
   * @async
   * @param {string} address - The block address of the record version `feedkey@seq`
   *    where `feedkey` is the hex-encoded public key of a feed and `seq` is a sequence number (uint).
   */
  async getVersion (address) {
    const [key, seq] = address.split('@');
    const version = await this.fetch(`/db/${key}/${seq}`);
    return this.store.cacheRecord(version)
  }

  /**
   * Deletes a record.
   *
   * @async
   * @param {object} record - The record to delete. Has to have `{ id, type }` properties set.
   * @return {Promise<object>} - An object with `{ id, type }` properties of the deleted record.
   */
  async del (record) {
    return this.fetch(`/record/${record.type}/${record.id}`, {
      method: 'DELETE'
    })
  }

  /**
   * Add a new type to the collection.
   *
   * @async
   * @param {object} schema - A schema object.
   * @throws Throws if the schema object is invalid or cannot be saved.
   * @return {Promise<object>} A promise that resolves to the saved schema object.
   */
  async putType (spec) {
    const type = this.schema.addType(spec);
    return this.fetch('/schema', {
      method: 'POST',
      body: type.toJSON()
    })
  }

  /**
   * Create a writable stream to put records into the collection.
   *
   * Example:
   * ```javascript
   * const batchStream = collection.createBatchStream()
   * batch.write(record)
   * batch.close()
   * ```
   *
   * @return {Writable<Record>} A writable stream
   */
  createBatchStream () {
    const self = this;
    const stream = new Transform({
      open (cb) {
        stream.finished = self.fetch('/', {
          method: 'POST',
          requestType: 'stream',
          headers: {
            'content-type': 'application/x-ndjson'
          },
          params: { batch: true },
          body: this
        }).catch(err => stream.destroy(err));
        cb();
      },
      transform (record, cb) {
        if (!record.id) record.id = uuid();
        record = self.schema.RecordVersion(record);
        // record = self.store.cacheRecord(record)
        const json = JSON.stringify(record);
        this.push(json + '\n');
        cb();
      }
    });
    return stream
  }

  /**
   * Wait for all pending indexing operations to be finished.
   *
   * @async
   * @return {Promise<void>}
   */
  async sync () {
    return this.fetch('/sync')
  }

  /**
   * Subscribe to events on this collection.
   *
   * Returns a readable stream that emits event objects.
   * They look like this:
   * `{ event: string, data: object }`
   *
   * Events are:
   *
   * * `update`: with data `{ lseq }`
   * * `feed`: with data `{ key }`
   * * `schema-update`
   *
   * @return {Readable<object>}
   */
  createEventStream () {
    if (!this._eventStream) {
      this._initEventSource();
    }
    return this._eventStream.subscribe()
  }

  async _initEventSource () {
    if (this._eventStream) return
    this._eventStream = new TeeStream();

    const onerror = _err => {
      // TODO: Where should the error go?
      // this.log.error('Error initializing event source: ' + err.message)
      // if (this._eventStream) this._eventStream.destroy(err)
      // if (this._eventStream) this._eventStream.close()
      this._eventStream = null;
    };

    try {
      if (!this.workspace.opened) await this.workspace.open();
    } catch (err) {
      onerror();
      return
    }

    this._eventSource = this.workspace.createEventSource('/events', {
      endpoint: this.endpoint,
      onerror,
      onmessage: (eventObject) => {
        this.log.trace('event: ' + eventObject.event);
        this._eventStream.write(eventObject);
        this.emit(eventObject.event, eventObject.data);
      }
    });
  }

  /**
   * Pull live updates from the server as they happen.
   *
   * After calling this method once, all new records and record versions
   * are pulled from the server once available. The `update` event
   * is emitted when new records are about to arrive.
   */
  pullLiveUpdates () {
    if (this._liveUpdates) return
    this._liveUpdates = true;
    const eventStream = this.createEventStream();
    eventStream.on('data', event => {
      if (event.event !== 'update') return
      const lseq = event.data.lseq;
      if (!lseq || lseq < this.length) return
      const oldLength = this.length;
      this._length = lseq + 1;
      for (let i = oldLength + 1; i < this.length; i++) {
        this.get({ lseq: i }).catch(e => {});
      }
      this.emit('update', lseq);
    });
  }

  /**
   * Subscribe to this collection.
   *
   * This will fetch all records from the first to the last and then waits for new records.
   * Currently only intended for usage in bots (not in short-running Browser clients).
   *
   * @todo: Prefix client ID to subscription name.
   * @todo: Allow to subscribe from now instead of from the beginning.
   *
   * @param {string} Subscription name. Has to be unique per running Sonar instance
   * @param {function} Async callback function that will be called for each incoming record
   */
  async subscribe (name, onRecord) {
    const self = this;
    await this._initEventSource();

    run().catch(err => {
      // TODO: How to deal with these errors?
      this.log.error('subscription error', err);
    });

    return true

    async function run () {
      // const eventStream = self.createEventStream()
      while (true) {
        // wait for one incoming update event
        const nextUpdatePromise = new Promise((resolve, reject) => {
          self.once('update', resolve);
        });

        const batch = await self._pullSubscription(name);
        for (const message of batch.messages) {
          try {
            const record = self.store.cacheRecord(message);
            await onRecord(record);
            // TODO: Do we want to ack for each message or for each batch?
            // Likely let the subscriber decide.
            await self._ackSubscription(name, message.lseq);
          } catch (err) {
            // TODO: What to do with errors here?
            self.log.error(`Error in subscription ${name}: ${err.message}`);
            debug$1(err);
            // TODO: This will easily cause an endless loop becaues it'll
            // just keep refreshing and running into errors ;-)
          }
        }
        if (batch.finished) {
          await nextUpdatePromise;
        }
      }
    }
  }

  async _pullSubscription (name, opts) {
    // TODO: Prefix name with client id.
    return this.fetch('/subscription/' + name, {
      query: opts
    })
  }

  async _ackSubscription (name, cursor) {
    // TODO: Prefix name with client id.
    return this.fetch('/subscription/' + name + '/' + cursor, {
      method: 'post'
    })
  }

  /**
   * Reindex the secondary indexes (views) in this collection.
   *
   * Use with care, this can be expensive.
   *
   * @param {Array<string>} Optional array of view names to reindex.
   *  If unset all views will be reindexed.
   */
  async reindex (views) {
    const params = {};
    if (views) params.views = views;
    return this.fetch('/reindex', {
      method: 'post',
      params
    })
  }

  async fetch (path, opts = {}) {
    if (!opts.endpoint) opts.endpoint = this.endpoint;
    return this.workspace.fetch(path, opts)
  }
}

class TeeStream extends Writable {
  constructor (opts) {
    super(opts);
    this._streams = new Set();
  }

  subscribe (opts) {
    const stream = new Readable(opts);
    stream.on('destroy', () => this._streams.remove(stream));
    this._streams.add(stream);
    return stream
  }

  _write (data, cb) {
    for (const stream of this._streams) {
      stream.push(data);
    }
    cb();
  }

  _destroy (err) {
    for (const stream of this._streams) {
      stream.destroy(err);
    }
    super.destroy(err);
  }
}

var collection = Collection;

const debug = require$$0('sonar:bots');

const SESSION_HEADER = 'X-Sonar-Bot-Session-Id';

class Bots {
  constructor (client) {
    this.client = client;
    this.bots = new Map();
    this.log = this.client.log.child({ name: 'bots' });
  }

  async close () {
    this._eventSource.close();
  }

  async join (botName, collection) {
    const res = await this.fetch('/join', {
      method: 'POST',
      body: { bot: botName, collection }
    });
    return res
  }

  async leave (botName, collection) {
    const res = await this.fetch('/leave', {
      method: 'POST',
      body: { bot: botName, collection }
    });
    return res
  }

  async info () {
    const res = await this.fetch('/info');
    return res
  }

  async command (botName, command, args, env) {
    const path = '/command';
    const res = await this.fetch(path, {
      method: 'POST',
      body: { bot: botName, command, args, env }
    });
    return res
  }

  async commandStatus (botName, requestId) {
    const path = `/status/${botName}/${requestId}`;
    const res = await this.fetch(path);
    return res
  }

  async getCommands () {
  }

  async _initListener () {
    this._init = true;

    // const path = this.client.endpoint + '/bot/events'
    // const headers = {
    //   ...this.client.getAuthHeaders(),
    //   ...this.getHeaders()
    // }

    const path = '/bot/events';
    this._eventSource = this.client.createEventSource(path, {
      headers: this.getHeaders(),
      onmessage: this._onmessage.bind(this),
      onerror: err => {
        // TODO: Where do these errors go?
        // TODO: After a couple of fails die.
        this.log.error({ err, message: 'Event source error' });
      }
    });
    // this._eventSource = new EventSource(path, { headers })
    // this._eventSource.addEventListener('message', message => {
    //   try {
    //     const event = JSON.parse(message.data)
    //     this._onmessage(event)
    //   } catch (e) {}
    // })
    // this._eventSource.addEventListener('error', err => {
    //   // TODO: Where do these errors go?
    //   // TODO: After a couple of fails die.
    //   this.log.error({ err, message: 'Event source error' })
    // })
  }

  async reply ({ requestId, error, result }) {
    await this.fetch('/reply', {
      method: 'POST',
      body: {
        requestId,
        error,
        result
      }
    });
  }

  async _onmessage (message) {
    try {
      const { bot: name, op, data, requestId } = message;
      const bot = this.bots.get(name);
      if (!bot) throw new Error('Unknown bot: ' + bot)
      try {
        if (op === 'join') {
          const { collection: collectionKey } = data;
          const collection = await this.client.openCollection(collectionKey);
          await bot.onjoin(collection, data.config);
          await this.reply({ requestId });
        } else if (op === 'leave') {
          const { collection: collectionKey } = data;
          const collection = await this.client.openCollection(collectionKey);
          await bot.onleave(collection);
          await this.reply({ requestId });
        } else if (op === 'command') {
          const { command, args, env } = data;
          if (env.collection) {
            env.collection = await this.client.openCollection(env.collection);
          }
          const result = await bot.oncommand(command, args, env);
          await this.reply({ requestId, result });
        }
      } catch (err) {
        this.log.error({ message: 'bot onmessage handle error: ' + err.message + ' from ' + JSON.stringify(message), err });
        debug(err);
        await this.reply({
          requestId: requestId,
          error: err.message
        });
      }
    } catch (err) {
      // TODO: Where to these errors go?
      this.log.error({ message: 'bot onmessage error', err });
    }
  }

  getHeaders () {
    const headers = {};
    if (this._sessionId) {
      headers[SESSION_HEADER] = this._sessionId;
    }
    return headers
  }

  async fetch (url, opts = {}) {
    url = '/bot' + url;
    opts.headers = opts.headers || {};
    opts.headers = { ...this.getHeaders(), ...opts.headers };
    return this.client.fetch(url, opts)
  }

  async register (name, spec, handlers) {
    if (typeof spec !== 'object') throw new Error('Spec must be an object')
    spec.name = name;
    const { config, sessionId } = await this.fetch('/register', {
      method: 'POST',
      body: spec
    });
    this._sessionId = sessionId;
    if (!this._init) await this._initListener();
    this.bots.set(name, new Bot({
      spec,
      client: this.client,
      config,
      handlers
    }));
  }

  // async configure (name, config) {
  //   const config = await this.client.fetch('/bot/configure/' + name, config)
  // }
}

class Bot {
  constructor ({ spec, config, handlers, client }) {
    this.spec = spec;
    this.name = spec.name;
    this.client = client;
    this.config = config;
    if (typeof handlers === 'function') {
      handlers = handlers({ spec, config, client });
    }
    this.handlers = handlers;
    this.sessions = new Map();
    this.opened = false;
    this.log = this.client.log.child({ name: 'bot:' + this.name });
  }

  async open () {
    if (this.opened) return
    if (this.opening) return this.opening
    let _resolve;
    this.opening = new Promise(resolve => (_resolve = resolve));
    if (this.handlers.open) await this.handlers.open();
    _resolve();
    this.opened = true;
    this.log.debug('open');
  }

  async onjoin (collection, config) {
    await this._ensureTypes(collection);
    await this.open();
    if (!this.opened && this.handlers.open) {
      await this.handlers.open();
      this.opened = true;
    }
    const session = await this.handlers.onjoin(collection, config);
    if (session.open) {
      await session.open();
    }
    if (session.onrecord) {
      collection.subscribe('bot:' + this.name, session.onrecord.bind(session));
    }
    this.sessions.set(collection.key, session);
    this.log.debug({ message: 'join', collection });
  }

  async _ensureTypes (collection) {
    if (!this.spec.types) return
    for (const typeSpec of this.spec.types) {
      if (!collection.schema.hasType(typeSpec)) {
        await collection.putType(typeSpec);
      }
    }
  }

  async onleave (collection) {
    const session = this.sessions.get(collection.key);
    if (!session) return
    if (session.close) await session.close();
    this.sessions.delete(collection.key);
    this.log.debug({ message: 'leave', collection });
  }

  async oncommand (command, args, env) {
    await this.open();

    if (!env.collection && this.handlers.oncommand) {
      this.log.debug('workspace command: ' + command);
      return await this.handlers.oncommand(command, args)
    }

    this.log.debug({ message: 'collection command: ' + command, collection: env.collection });
    const session = this.sessions.get(env.collection.key);
    if (!session) throw new Error('Bot did not join collection')
    if (!session.oncommand) throw new Error('Bot cannot handle commands')

    return await session.oncommand(command, args)
  }
}

var bots = Bots;
var Bots_1 = Bots;
bots.Bots = Bots_1;

const { EventEmitter } = require$$0$1;





const { Logger } = common;

const {
  DEFAULT_ENDPOINT,
  DEFAULT_WORKSPACE
} = constants;

function defaultWorkspace () {
  return (process && {} && {}.WORKSPACE) || DEFAULT_WORKSPACE
}

class Workspace extends EventEmitter {
  /**
   * A Sonar workspace. Provides methods to open collection under this endpoint.
   *
   * @constructor
   * @param {object} [opts] - Optional options.
   * @param {string} [opts.url=http://localhost:9191/api/v1/default] - The API endpoint to talk to.
   * @param {string} [opts.accessCode] - An access code to login at the endpoint.
   * @param {string} [opts.token] - A JSON web token to authorize to the endpoint.
   * @param {string} [opts.name] - The name of this client.
   */
  constructor (opts = {}) {
    super();
    if (opts.endpoint) {
      const workspace = opts.workspace || defaultWorkspace();
      const endpoint = opts.endpoint || DEFAULT_ENDPOINT;
      this.endpoint = `${endpoint}/${workspace}`;
    } else if (opts.url) {
      this.endpoint = opts.url;
      // TODO: deprecate, ue only URL.
    } else {
      this.endpoint = DEFAULT_ENDPOINT;
    }
    if (this.endpoint.endsWith('/')) {
      this.endpoint = this.endpoint.substring(0, this.endpoint.length - 1);
    }

    this._collections = new Map();
    this._id = opts.id || randomBytes(16).toString('hex');
    this._token = opts.token;
    this._accessCode = opts.accessCode;
    this._eventSources = [];

    this.log = opts.log || new Logger();
    this.bots = new bots(this);
  }

  get token () {
    return this._token
  }

  /**
   * Closes the client.
   *
   * @async
   * @return {Promise<void>}
   */
  async close () {
    for (const eventSource of this._eventSources) {
      eventSource.close();
    }
    for (const collection of this._collections.values()) {
      collection.close();
    }
  }

  async open () {
    if (this.opened) return
    if (!this._openPromise) this._openPromise = this._open();
    await this._openPromise;
  }

  async _open () {
    await this._login();
    this.opened = true;
  }

  // TODO: Support re-logins
  async _login () {
    if (this._accessCode && !this._token) {
      const res = await this.fetch('/login', { params: { code: this._accessCode }, method: 'POST', opening: true });
      const token = res.token;
      this._token = token;
    }
  }

  /**
   * Get a list of all collections available on this endpoint.
   *
   * @async
   * @return {Promise.<object[]>} Promise that resolves to an array of collection info objects.
   */
  async listCollections () {
    const info = await this.fetch('/');
    return info.collections
  }

  /**
   * Creates a collection with name name on the Sonar server. The name may not contain whitespaces. opts is an optional object with:
   *
   * @async
   * @param {string} name - Name of the new collection, may not contain whitespaces.
   * @param {object} [opts] - Optional options object.
   * @param {string} [opts.key] - Hex string of an existing collection. Will then sync this collection instead of creating a new, empty collection.
   * @param {string} [opts.alias] - When setting key, alias is required and is your nick name within this collection.
   * @return {Promise<Collection>} The created collection.
   */
  async createCollection (name, opts = {}) {
    opts.name = name;
    await this.fetch('/collection', {
      method: 'POST',
      body: opts
    });
    const collection = await this.openCollection(name, { reset: true });
    return collection
  }

  // TODO: Move to Collection.update()?
  // TODO: info == config?
  /**
   * Updates the config of a collection.
   *
   * @async
   * @param {string} name - Name of the collection.
   * @param {object} info - [TODO:description]
   * @param {boolean} info.share - Controls whether a collection is shared via p2p.
   * @return {Promise<void>}
   */
  async updateCollection (name, info) {
    return this.fetch('/collection/' + name, {
      method: 'PATCH',
      body: info
    })
  }

  /**
   * Returns a Collection object for a given key or name of a collection.
   *
   * @async
   * @param {string} keyOrName - Key or name of the collection to open/return.
   * @return {Promise<Collection>}
   */
  async openCollection (keyOrName, opts = {}) {
    if (this._collections.has(keyOrName)) {
      const collection = this._collections.get(keyOrName);
      if (!collection.opened) await collection.open(opts.reset);
      return collection
    }

    const collection$1 = new collection(this, keyOrName);
    collection$1.on('open', () => this.emit('collection-open', collection$1));
    this._collections.set(keyOrName, collection$1);
    // This will throw if the collection does not exist.
    try {
      await collection$1.open(opts.reset);
      this._collections.set(collection$1.name, collection$1);
      this._collections.set(collection$1.key, collection$1);
      return collection$1
    } catch (err) {
      this._collections.delete(keyOrName);
      throw err
    }
  }

  getCollection (keyOrName) {
    return this._collections.get(keyOrName)
  }

  getAuthHeaders (opts = {}) {
    const headers = {};
    const token = this._token || opts.token;
    if (token) {
      headers.authorization = 'Bearer ' + token;
    }
    return headers
  }

  getHeaders (opts = {}) {
    return this.getAuthHeaders()
  }

  /**
   * Fetch a resource.
   *
   * This is a wrapper around the fetch web API. It should be API compatible to fetch,
   * with the following changes:
   *
   * @async
   * @param {string} [opts.requestType='json'] Request encoding and content type.
   *   Supported values are 'json' and 'binary'
   * @param {string} [opts.responseType='text'] Response encoding. If the response
   *    has a JSON content type, will always be set to 'json'.
   *    Supported values are 'text', 'binary' and 'stream'.
   * @param {object} [opts.params] Query string parameters (will be encoded correctly).
   *
   * @return {Promise<object>} If the response has a JSON content type header, the
   *    decoded JSON will be returned. if opts.responseType is 'binary' or 'text',
   *    the response will be returned as a buffer or text.
   *
   * TODO: Rethink the default responseType cascade.
   */
  async fetch (url, opts = {}) {
    if (!this.opened && !opts.opening) {
      await this.open();
    }
    if (!opts.endpoint && this.endpoint) {
      opts.endpoint = this.endpoint;
    }
    if (!opts.headers) opts.headers = {};
    opts.headers = {
      ...opts.headers,
      ...this.getAuthHeaders(opts)
    };
    opts.log = message => this.log.debug({ message, name: 'fetch' });
    return fetch_1(url, opts)
  }

  createEventSource (path, opts = {}) {
    if (!opts.endpoint && this.endpoint) opts.endpoint = this.endpoint;
    opts.headers = Object.assign(opts.headers || {}, this.getHeaders(opts));
    let url = (opts.endpoint || '') + path;
    url = new URL(url);
    if (this._token || opts.token) {
      url.searchParams.set('token', this._token || opts.token);
    }
    const eventSource = new EventSource(url.toString(), opts);
    eventSource.addEventListener('message', message => {
      try {
        const event = JSON.parse(message.data);
        if (opts.onmessage) opts.onmessage(event);
      } catch (e) {}
    });
    eventSource.addEventListener('error', err => {
      // TODO: Where do these errors go?
      // TODO: After a couple of fails die.
      if (opts.onerror) opts.onerror(err);
      else console.error('Event source error', err);
    });
    this._eventSources.push(eventSource);
    return eventSource
  }
}

var workspace = Workspace;

var searchquerybuilder = class SearchQueryBuilder {
  constructor (schema) {
    this.schema = schema;
    this.query = {};
  }

  bool (boolType, queries) {
    if (!this.query.hasOwnProperty('bool')) {
      this.query.bool = {};
    }
    this.query.bool[boolType] = queries;
    return this
  }

  phrase (field, terms) {
    if (!this.query.hasOwnProperty('phrase')) {
      this.query.phrase = {};
    }
    this.query.phrase[field] = { terms };
    return this
  }

  limit (limit) {
    this.limit = limit;
    return this
  }

  term (field, value) {
    return { term: { [field]: value } }
  }

  getQuery () {
    return { query: this.query, limit: this.limit }
  }
};

const { DEFAULT_COLLECTION } = constants;


var legacyClient = class LegacyClient extends workspace {
  constructor (opts) {
    super(opts);
    this._focus = null;
    this._defaultCollection = opts.collection || DEFAULT_COLLECTION;
  }

  focusCollection (name) {
    this._focus = this.openCollection(name);
  }

  async focusedCollection () {
    if (!this._focus && this._defaultCollection) this.focusCollection(this._defaultCollection);
    this._focusedCollection = await this._focus;
    return this._focus
  }

  // TODO: Remove.
  async getCurrentCollection () {
    return this.focusedCollection()
  }

  async info () {
    return this.fetch('/')
  }

  async getTypes () {
    const collection = await this.focusedCollection();
    return collection.schema.getTypes()
  }

  async getType (name) {
    const collection = await this.focusedCollection();
    return collection.schema.getType(name)
  }

  async putType (name, schema) {
    schema.name = name;
    const collection = await this.focusedCollection();
    return collection.putType(schema)
  }

  // TODO: Remove.
  getSchemas () {
    return this.getTypes()
  }

  getSchema (name) {
    return this.getType(name)
  }

  async putFeed (key, opts) {
    const collection = await this.focusedCollection();
    return collection.putFeed(key, opts)
  }

  async get (req, opts) {
    const collection = await this.focusedCollection();
    return collection.get(req, opts)
  }

  async put (record) {
    const collection = await this.focusedCollection();
    return collection.put(record)
  }

  async del (record) {
    const collection = await this.focusedCollection();
    return collection.del(record)
  }

  async sync (view) {
    const collection = await this.focusedCollection();
    return collection.sync(view)
  }

  async query (name, args, opts) {
    const collection = await this.focusedCollection();
    return collection.query(name, args, opts)
  }

  async search (args, opts) {
    if (typeof args === 'string') {
      args = JSON.stringify(args);
    } else if (args instanceof searchquerybuilder) {
      args = args.getQuery();
    }
    const collection = await this.focusedCollection();
    return collection.query('search', args, opts)
  }

  // FS methods
  async getDrives () {
    const collection = await this.focusedCollection();
    return collection.fs.listDrives()
  }

  async readdir (path) {
    const collection = await this.focusedCollection();
    return collection.fs.readdir(path)
  }

  async writeFile (path, file, opts) {
    const collection = await this.focusedCollection();
    return collection.fs.writeFile(path, file, opts)
  }

  async readFile (path, opts) {
    const collection = await this.focusedCollection();
    return collection.fs.readFile(path, opts)
  }

  async statFile (path) {
    const collection = await this.focusedCollection();
    return collection.fs.stat(path)
  }

  // Resources
  // TODO: Rethink the resource API and model
  async createResource (value, opts) {
    const collection = await this.focusedCollection();
    return collection.resources.create(value, opts)
  }

  async readResourceFile (record, opts) {
    const collection = await this.focusedCollection();
    return collection.resources.readFile(record, opts)
  }

  async writeResourceFile (record, file, opts) {
    const collection = await this.focusedCollection();
    return collection.resources.writeFile(record, file, opts)
  }

  async resolveResourceURL (record) {
    const collection = await this.focusedCollection();
    return collection.resources.resolveFileURL(record)
  }

  // Subscriptions
  async pullSubscription (name, opts) {
    const collection = await this.focusedCollection();
    return collection.pullSubscription(name, opts)
  }

  async ackSubscription (name, cursor) {
    const collection = await this.focusedCollection();
    return collection.ackSubscription(name, cursor)
  }
};

var client = legacyClient;
var Workspace_1 = workspace;
var SearchQueryBuilder_1 = searchquerybuilder;
client.Workspace = Workspace_1;
client.SearchQueryBuilder = SearchQueryBuilder_1;

export default client;
export { SearchQueryBuilder_1 as SearchQueryBuilder, Workspace_1 as Workspace };
//# sourceMappingURL=index.js.map
